import re
import xxhash
import inflect

INFLECTOR = inflect.engine()
PLURAL_DICT = {'spectrum': 'spectra', 'noss': 'nosses', 'use': 'uses'}
exceptions = {'1_D': '1D', '2_D': '2D', '3_D': '3D'}

def snakecase2camelcase(name):
    """
    Returns a string with underscores between words in camelcase name
    """
    return ''.join(x.capitalize() for x in name.split('_'))


def camelcase2snakecase(name):
    """
    Returns a string with underscores between words in camelcase name
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)
    for a, b in exceptions.items():
        name = name.replace(a, b)
    return name.lower()

def make_plural(name):
    if name.endswith('_group'):
        raise ValueError(f"Cannot pluralise {name} since it ends with '_group'")
    for k, v in PLURAL_DICT.items():
        if name.endswith(k):
            return name[:-len(k)] + v
        if name.endswith(v):
            return name + '_group'
    plural_form = INFLECTOR.plural_noun(name)
    single_form = INFLECTOR.singular_noun(name)  # returns False when name is already singular
    if not single_form:
        return plural_form
    return name + '_group'

def make_singular(name):
    if name.endswith('_group'):
        return name[:-len('_group')]
    for v, k in PLURAL_DICT.items():
        if name.endswith(k):
            return name[:-len(k)] + v
        if name.endswith(v):
            return name
    single_form = INFLECTOR.singular_noun(name)  # returns False when name is already singular
    if not single_form:
        return name
    return single_form

class Varname:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

def quote(x):
    """
    Return a quoted string if x is a string otherwise return x
    """
    if isinstance(x, str):
        return f"'{x}'"
    if isinstance(x, Varname):
        return x.name
    return x


def hash_pandas_dataframe(df):
    digester = xxhash.xxh64()
    for i in df.astype(str).values.ravel():
        digester.update(i)
    return digester.hexdigest()


def int_or_none(x):
    if x is None:
        return None
    return int(x)


import textwrap

import networkx as nx

from weaveio.basequery.factor import FactorFrozenQuery
from weaveio.hierarchy import Hierarchy, Graphable, GraphableMeta


def _convert_obj(obj, data=None):
    if isinstance(obj, str):
        obj = obj.lower()
        try:
            obj = data.singular_hierarchies[obj]
        except KeyError:
            obj = data.plural_hierarchies[obj]
    elif isinstance(obj, (Graphable, GraphableMeta)):
        obj = obj
    elif hasattr(obj, 'handler'):
        data = obj.handler.data
        obj = obj.hierarchy_type
    else:
        raise TypeError(f"{obj} is not a recognised type of object. Use a name (string), a query, or the type directly")
    return obj, data


def attributes(obj, data=None):
    obj, data = _convert_obj(obj, data)
    return sorted(set(obj.products_and_factors))


def objects(obj, data=None):
    obj, data = _convert_obj(obj, data)
    neighbors = set(data.relation_graph.predecessors(obj)) | set(data.relation_graph.successors(obj))
    relations = []
    for b in neighbors:
        try:
            data.find_hierarchy_paths(obj, b, plural=False)
        except (nx.NetworkXNoPath, KeyError):
            relations.append(b.plural_name)
        else:
            relations.append(b.singular_name)
    relations.sort()
    return relations


def explain_object(obj, data=None):
    obj, data = _convert_obj(obj, data)
    objs = objects(obj, data)
    attrs = attributes(obj, data)
    print(f"{obj.singular_name:=^40s}")
    if obj.__doc__:
        print('\n'.join(textwrap.wrap(textwrap.dedent('\n'.join(obj.__doc__.split('\n'))))))
        print()
    if obj.__bases__ != (Hierarchy, ):
        print(f"A {obj.singular_name} is a type of {obj.__bases__[0].singular_name}")
    if obj.idname is not None:
        print(f"A {obj.singular_name} has a unique id called '{obj.idname}'")
    else:
        print(f"A {obj.singular_name} has no unique id that can be used")
    if obj.identifier_builder:
        print(f"{obj.plural_name} are identified by {tuple(obj.identifier_builder)}")
    print(f'one {obj.singular_name} is linked to:')
    for o in objs:
        if data.is_plural_name(o):
            print('\t- many', o)
        else:
            print('\t- 1', o)
    print(f'a {obj.singular_name} directly owns these attributes:')
    for a in attrs:
        print('\t-', a)



def explain_factor(factor, data = None):
    if isinstance(factor, FactorFrozenQuery):
        for f in factor.factors:
            explain_factor(f, factor.data)
    else:
        factor = data.singular_name(factor)
        hierarchies = {h for h in data.factor_hierarchies[factor] if not h.is_template}
        if len(hierarchies) > 1:
            print(f'{factor}s are owned by multiple different objects ({[h.singular_name for h in hierarchies]}).'
                  f'\nThey could be entirely different things.')
            print(f'You will need to specify one of the parent objects below for {factor} when querying.')
        for h in hierarchies:
            if factor in h.products:
                print(f'A {factor} is a product (binary data kept out-of-database) of a {h.singular_name}')
            elif factor == h.idname:
                print(f'{factor} is the unique id name of a {h.singular_name}')
            else:
                print(f"{factor} is an attribute belonging to {h.singular_name}")


def explain_relation(a, b, data = None):
    a, data = _convert_obj(a, data)
    try:
        b, data = _convert_obj(b, data)
        try:
            data.find_hierarchy_paths(a, b, plural=False)
        except nx.NetworkXNoPath:
            print(f"- A {a.singular_name} has many {b.plural_name}")
        else:
            print(f"- A {a.singular_name} has only one {b.singular_name}")
    except (KeyError, TypeError):
        if isinstance(b, FactorFrozenQuery):
            data = b.data
            b = b.factors[0]
        hierarchies = ', '.join({i.singular_name for i in data.factor_hierarchies[b]})
        try:
            data.find_factor_paths(a, b, plural=False)
        except nx.NetworkXNoPath:
            print(f"- A {a.singular_name} has many {b}s (belonging to {hierarchies})")
        else:
            print(f"- A {a.singular_name} has 1 {b} (belonging to {hierarchies})")


def explain(a, b=None, data=None):
    try:
        explain_object(a, data)
    except (KeyError, AttributeError):
        if data is None:
            raise ValueError(f"You must supply `data=data` to `explain` when asking about a string name")
        explain_factor(a, data)
    if b is not None:
        try:
            explain_object(b, data)
        except (KeyError, AttributeError):
            if data is None:
                raise ValueError(f"You must supply `data=data` to `explain` when asking about a string name")
            explain_factor(b, data)
        print('='*40)
        explain_relation(a, b, data)
        explain_relation(b, a, data)
    print('='*40)



"""
Tests to ensure that
Query(predicates, return_properties, return_nodes, conditions, exists)
are written to neo4j correctly.e

I.e.
"""
from textwrap import dedent

import pytest
from weaveio.basequery.query import FullQuery
from weaveio.basequery.query_objects import Collection, Path, Generator, Condition


@pytest.fixture(scope='function')
def g():
    return Generator()


def test_matches_must_overlap(g):
    a, b, c, d = g.nodes('A', 'B', 'C', 'D')
    path1 = Path(a, '->', b)
    path2 = Path(c, '->', d)
    with pytest.raises(ValueError):
        FullQuery([path1, path2])


def test_cannot_return_nodes_that_dont_end_a_path(g):
    a, b, c, d, e, f = g.nodes('A', 'B', 'C', 'D', 'E', 'F')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    with pytest.raises(KeyError):
        FullQuery([path1, path2], returns=[a])
    with pytest.raises(KeyError):
        FullQuery([path1, path2], returns=[a.property_a])


def test_current_node(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2])
    assert query.current_node == c


def test_matches_only(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2])
    q, _ = query.to_neo4j()
    assert q == f'MATCH (a0:A)->(b0:B)\nMATCH (b0)->(c0:C)\nWITH b0, c0\nRETURN'


def test_return_nodes(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2], returns=[c, b])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0
    RETURN c0, b0
    """).strip('\n')
    assert q == expected


def test_return_properties(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2], returns=[c.property_c, b.property_b])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0, c0.property_c as c0_property_c, b0.property_b as b0_property_b
    RETURN c0_property_c, b0_property_b
    """).strip('\n')
    assert q == expected


def test_return_nodes_properties(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2], returns=[c.property_c, b.property_b, b, c])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0, c0.property_c as c0_property_c, b0.property_b as b0_property_b
    RETURN c0_property_c, b0_property_b, b0, c0
    """).strip('\n')
    assert q == expected


def test_condition_on_property(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    condition = Condition(a.id, '=', 'idvalue')
    query = FullQuery([path1, path2], conditions=condition)
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WHERE (a0.id = 'idvalue')
    WITH b0, c0
    RETURN
    """).strip('\n')
    assert q == expected


def test_property_branches_add_with_statements(g):
    a, b, c, d, e, f = g.nodes('A', 'B', 'C', 'D', 'E', 'F')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    branch1 = Path(c, '->', d)
    branch2 = Path(c, '->', e, '->', f)
    query = FullQuery([path1, path2],
                      branches={branch1: [d.property_d], branch2: [f.property_f]},
                      returns=[c, b.property_b, d.property_d, f.property_f])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0, b0.property_b as b0_property_b
    OPTIONAL MATCH (c0)->(d0:D)
    WITH b0, c0, b0_property_b, d0.property_d as d0_property_d
    OPTIONAL MATCH (c0)->(e0:E)->(f0:F)
    WITH b0, c0, b0_property_b, d0_property_d, f0.property_f as f0_property_f
    RETURN c0, b0_property_b, d0_property_d, f0_property_f
    """).strip('\n')
    assert q == expected


def test_collect_property_branches_uses_collection(g):
    a, b, c = g.nodes('A', 'B', 'C')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    query = FullQuery([path1, path2], returns=[Collection(b.property_b, 'property_b')])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0, collect(b0.property_b) as property_b
    RETURN property_b
    """).strip('\n')
    assert q == expected


def test_collect_property_of_root_uses_collection(g):
    a, b, c, d, e, f = g.nodes('A', 'B', 'C', 'D', 'E', 'F')
    path1 = Path(a, '->', b)
    path2 = Path(b, '->', c)
    branch1 = Path(c, '->', d)
    branch2 = Path(c, '->', e, '->', f)
    collect_f = Collection(f.property_f, 'property_f')
    query = FullQuery([path1, path2],
                      branches={branch1: [d.property_d], branch2: [collect_f]},
                      returns=[c, Collection(b.property_b, 'property_b'), d.property_d, collect_f])
    q, _ = query.to_neo4j()
    expected = dedent("""
    MATCH (a0:A)->(b0:B)
    MATCH (b0)->(c0:C)
    WITH b0, c0, collect(b0.property_b) as property_b
    OPTIONAL MATCH (c0)->(d0:D)
    WITH b0, c0, property_b, d0.property_d as d0_property_d
    OPTIONAL MATCH (c0)->(e0:E)->(f0:F)
    WITH b0, c0, property_b, d0_property_d, collect(f0.property_f) as property_f
    RETURN c0, property_b, d0_property_d, property_f
    """).strip('\n')
    assert q == expected



from .base import CypherQuery, Statement, CypherVariable, Collection


class Unwind(Statement):
    def __init__(self, *args: CypherVariable, enumerated=False):
        output_variables = [CypherVariable('unwound_' + a.namehint.replace('$', "")) for a in args]
        super(Unwind, self).__init__(args, output_variables)
        self.passed_outputs = self.output_variables.copy()
        if enumerated or len(self.input_variables) > 1:
            self.indexer_variable = CypherVariable('i')
            self.output_variables.append(self.indexer_variable)
            if enumerated:
                self.passed_outputs.append(self.indexer_variable)
        self.enumerated = enumerated

    def to_cypher(self):
        if len(self.input_variables) == 1 and not self.enumerated:
            return f"WITH * UNWIND {self.input_variables[0]} as {self.output_variables[0]}"
        else:
            outaliasstr = 'WITH *'
            for n, (i, o) in enumerate(zip(self.input_variables, self.output_variables[:-1])):
                outaliasstr += f', {i}[{self.indexer_variable}] as {o}'
            inliststr = f'[{",".join(map(str, self.input_variables))}]'
            return f"WITH *, apoc.coll.max([x in {inliststr} | SIZE(x)])-1 as m\n" \
                   f"UNWIND range(0, m) as {self.indexer_variable} {outaliasstr}"


class Collect(Statement):
    def __init__(self, previous, *args):
        self.previous = previous
        super(Collect, self).__init__(args, [Collection(a.namehint + 's') for a in args])

    def to_cypher(self):
        collections = ','.join([f'collect({a}) as {b}' for a, b in zip(self.input_variables, self.output_variables)])
        r = f"WITH " + ', '.join(map(str, self.previous))
        if len(collections):
            r += ', ' + collections
        return r

    def __getitem__(self, inarg):
        i = self.input_variables.index(inarg)
        try:
            return self.output_variables[i]
        except IndexError:
            raise KeyError(f"Cannot access {inarg} in this context, have you unwound it previously or left a WITH context?")

    def __delitem__(self, key):
        i = self.input_variables.index(key)
        del self.input_variables[i]


class NodeMap(CypherVariable):
    def __getitem__(self, item):
        query = CypherQuery.get_context()  # type: CypherQuery
        statement = GetItemStatement(self, item)
        query.add_statement(statement)
        return statement.out


class GetItemStatement(Statement):
    def __init__(self, mapping, item):
        self.mapping = mapping
        self.item = item
        ins = [mapping]
        if isinstance(item, CypherVariable):
            ins.append(item)
        itemname = getattr(item, 'namehint', 'item') if not isinstance(item, str) else item
        self.out = CypherVariable(mapping.namehint + '_' + itemname)
        self.temp = CypherVariable('temp')
        super(GetItemStatement, self).__init__(ins, [self.out], [self.temp])

    def to_cypher(self):
        if isinstance(self.item, CypherVariable):
            item = f'{self.item}'
        else:
            item = f"'{self.item}'"
        convert = f"WITH *, CASE WHEN apoc.meta.type({item}) = 'STRING' THEN {item} ELSE toString(toFloat({item})) END as {self.temp}"
        result = f'WITH *, {self.mapping}[{self.temp}] as {self.out}'
        return '\n'.join([convert, result])


class Copy(Statement):
    def __init__(self, *args):
        outs = [CypherVariable(a.namehint + '_copy') for a in args]
        super(Copy, self).__init__(args, outs)

    def to_cypher(self):
        eq = [f'{a} as {b}' for a, b in zip(self.input_variables, self.output_variables)]
        return f"WITH *, {', '.join(eq)}"

class GroupBy(Statement):
    def __init__(self, nodelist, propertyname):
        super(GroupBy, self).__init__([nodelist], [NodeMap(propertyname+'_dict')])
        self.propertyname = propertyname

    def to_cypher(self):
        return f"WITH *, apoc.map.groupBy({self.input_variables[0]}, '{self.propertyname}') as {self.output_variables[0]}"


import math
from itertools import zip_longest

import networkx as nx

def shortest_simple_paths(graph, source, target, weight=None):
    try:
        yield from nx.shortest_simple_paths(graph, source, target, weight)
    except nx.NetworkXNoPath:
        yield []


def _find_singular_simple_hierarchy_path(graph, start, end, optionals=False):
    """
    Assuming the graph contains only edges from top to bottom which are labelled singular and not optional.
    x-[singular]->y "y has 1 x", "x can have many y"
    In this case, we always look for start<-end first otherwise look for end<-start (and reverse the path)
    """
    def filt(u, v):
        edge = graph.edges[u, v]
        return edge['singular'] and (not edge['optional'] if optionals else True)  # this only works if all paths have equal weight otherwise
    g = nx.subgraph_view(graph, filter_edge=filt)
    try:
        path = next(nx.shortest_simple_paths(g, end, start, 'optional'))[::-1]
    except nx.NetworkXNoPath:
        path = next(nx.shortest_simple_paths(g, start, end, 'optional'))
    return path


def find_singular_simple_hierarchy_path(graph, start, end):
    """
    If there is a path without optionals then take it, otherwise we want the fewest optional edges
    """
    return _find_singular_simple_hierarchy_path(graph, start, end, optionals=False)



def _find_path(graph, a, b, force_single):
    singles = nx.subgraph_view(graph, filter_edge=lambda u, v: graph.edges[u, v]['singular'])
    # first try to find a singular path
    single_paths = [i for i in [next(shortest_simple_paths(singles, a, b)), next(shortest_simple_paths(singles, b, a))[::-1]] if i]
    if single_paths:
        return min(single_paths, key=len)
    elif force_single:
        raise nx.NetworkXNoPath('No path found between {} and {}'.format(a, b))
    # then try to find a non-singular path with one direction, either -> or <-
    # we dont want ()->()<-() when it's multiple
    restricted = nx.subgraph_view(graph, filter_edge=lambda u, v: graph.edges[u, v]['actual_number'] > 0)
    forward = next(shortest_simple_paths(restricted, a, b, 'actual_number'))
    forward_weight = nx.path_weight(restricted, forward, 'actual_number') or math.inf
    reversed = restricted.reverse()
    backward = next(shortest_simple_paths(reversed, a, b, 'actual_number'))
    backward_weight = nx.path_weight(reversed, backward, 'actual_number') or math.inf
    if not forward and not backward:
        raise nx.NetworkXNoPath('No path found between {} and {}'.format(a, b))
    if backward_weight < forward_weight:
        return backward
    return forward

def shortest_simple_paths_with_weight(graph, a, b, weight):
    for path in shortest_simple_paths(graph, a, b, weight):
        yield path, nx.path_weight(graph, path, weight)

def find_paths(graph, a, b, force_single):
    """
    generates all unidirectional paths from a->b or b<-a with the same shortest length.
    Conditions:
        1. Cannot mix -> and <-, since that would allow nonsense like l2single->aps<-l2stack
        2. Prefer paths with as few "optional" edges as possible
        3. Prefer paths that are single in some direction.
    In the case of generic traversals e.g. `ingested.camera`, there are multiple paths
    """
    # should decide how to include optional edges
    real = nx.subgraph_view(graph, filter_edge=lambda u, v: graph.edges[u, v]['actual_number'] > 0 and 'relation' in graph.edges[u, v])
    flipped = nx.subgraph_view(graph, filter_edge=lambda *e: e[::-1] in real.edges)
    real_paths = shortest_simple_paths_with_weight(real, a, b, weight='actual_number')
    flipped_paths = shortest_simple_paths_with_weight(flipped, a, b, weight='actual_number')
    paths = ((p, w) for paths in zip_longest(real_paths, flipped_paths, fillvalue=(None, None)) for p, w in paths if p)
    buffer = []
    buffer_weight = None
    for path, weight in paths:
        length = len(path)
        if length == buffer_weight or buffer_weight is None:
            buffer.append(path)
            buffer_weight = length
        elif length > buffer_weight:
            break
        elif length < buffer_weight:
            buffer = [path]
            buffer_weight = length
    if buffer:
        return buffer
    raise nx.NetworkXNoPath(f"No path found between {a} and {b}")
    # equivalent paths will be removed later at the arrow stage








def find_path(graph, a, b, force_single):
    paths = find_paths(graph, a, b, force_single)
    return paths[0]
    # nonoptional = nx.subgraph_view(graph, filter_edge=lambda *e: not graph.edges[e]['optional'])
    # try:
    #     path = _find_path(nonoptional, a, b, force_single)
    # except nx.NetworkXNoPath:
    #     path = _find_path(graph, a, b, force_single)
    # if len(path) == 1:
    #     return [path[0], path[0]]
    # return path


import logging
from collections import deque
from functools import reduce
from typing import Type, List, Union, Tuple, Dict, Set, Iterable

from py2neo import Relationship, Node, Subgraph
from py2neo.export import Table
from tqdm import tqdm

from .utilities import int_or_none
from .graph import Graph
from .hierarchy import Graphable, Hierarchy, Multiple, OneOf, Optional


class AttemptedSchemaViolation(Exception):
    pass


def get_all_class_bases(cls: Type[Graphable], exclude=Graphable) -> List[Type[Graphable]]:
    new = []
    for b in cls.__bases__:
        if b is exclude or not issubclass(b, exclude):
            continue
        new.append(b)
        new += get_all_class_bases(b, exclude=exclude)
    return new


def get_labels_of_schema_hierarchy(hierarchy: Union[Type[Hierarchy], Multiple], as_set=False):
    if isinstance(hierarchy, Multiple):
        hierarchy = hierarchy.node
    bases = get_all_class_bases(hierarchy)
    labels = [i.__name__ for i in bases]
    labels.append(hierarchy.__name__)
    labels.append('SchemaNode')
    if as_set:
        return frozenset(labels)
    return labels


def push_py2neo_schema_subgraph_cypher(subgraph: Subgraph) -> Tuple[str, Dict]:
    """
    Each node label is unique, so we just merge by label and then update properties
    for relationships we merge with empty relation and then update the properties
    """
    cypher = []
    params = {}
    for i, node in enumerate(subgraph.nodes):
        cypher.append(f"MERGE (n{i}{node.labels} {{name: '{node['name']}'}})")
        for k, v in node.items():
            if k != 'name':
                params[f'n{i}{k}'] = v
                cypher.append(f"SET n{i}.{k} = $n{i}{k}")
    for i, rel in enumerate(subgraph.relationships):
        a = list(subgraph.nodes).index(rel.start_node)
        b = list(subgraph.nodes).index(rel.end_node)
        rel_params = []
        for k, v in rel.items():
            params[f'r{i}{k}'] = v
            rel_params.append(f"{k}: $r{i}{k}")
        rel_params = ','.join(rel_params)
        cypher.append(f"MERGE (n{a})-[r{i}:{list(rel.types())[0]} {{{rel_params}}}]->(n{b})")
    return "\n".join(cypher), params


def get_all_schema_nodes(graph):
    return graph.execute("MATCH (n:SchemaNode) OPTIONAL MATCH (n)-[:IS_TYPE_OF]->(m) with n, collect(m.name) as bases return n.name, bases, properties(n)").to_table()

def get_all_schema_rows(graph, hierarchy_name) -> Table:
    cypher = f"""
    MATCH (n:{hierarchy_name}:SchemaNode {{name: '{hierarchy_name}'}})
    OPTIONAL MATCH (child:SchemaNode)<-[child_rel:HAS_CHILD]-(n) WHERE child_rel.originated = n.name
    with n, child, collect(child_rel) as child_rels
    OPTIONAL MATCH (parent:SchemaNode)-[parent_rel:IS_PARENT_OF]->(n) WHERE parent_rel.originated = n.name
    with n, child, child_rels, parent, collect(parent_rel) as parent_rels
    RETURN n, parent, child, parent_rels, child_rels
    """
    return graph.execute(cypher).to_table()

def hierarchy_dependency_tree(hierarchies, done=None):
    done = [] if done is None else done
    hierarchies = deque(sorted(list(hierarchies),
                               key=lambda x: len(x.children) + len(x.parents) + len(x.products)))
    nmisses = 0
    while hierarchies:
        if nmisses == len(hierarchies):
            raise TypeError(f"Dependency resolution impossible, proposed schema elements may have cyclic dependencies:"
                                           f"{list(hierarchies)}")
        hier = hierarchies.popleft()  # type: Type[Hierarchy]
        hier.instantate_nodes(hierarchies)
        if hier.is_template:
            continue  # don't need to add templates, they just provide labels
        dependencies = [d.node if isinstance(d, Multiple) else d for d in hier.parents + hier.children]
        dependencies = filter(lambda d: d != hier, dependencies)  # allow self-references
        if not all(d in done for d in dependencies):
            logging.info(f"{hier} put at the back because it requires dependencies which have not been written yet")
            hierarchies.append(hier)  # do it after the dependencies are done
            nmisses += 1
            continue
        yield hier
        nmisses = 0  # reset counter
        done.append(hier)


def hierarchy_type_tree(hierarchies: Iterable[Type[Hierarchy]], done: List[Type[Hierarchy]] = None):
    done = [] if done is None else done
    nmisses = 0
    hierarchies = deque(hierarchies)
    while hierarchies:
        if nmisses == len(hierarchies):
            raise TypeError(f"Type resolution impossible, proposed schema elements may have cyclic type dependencies:"
                                           f"{list(hierarchies)}")
        hier = hierarchies.popleft()  # type: Type[Hierarchy]
        dependencies = get_all_class_bases(hier, exclude=Hierarchy)
        if not all(d in done for d in dependencies):
            logging.info(f"{hier} put at the back because it requires dependencies which have not been written yet")
            hierarchies.append(hier)  # do it after the dependencies are done
            nmisses += 1
            continue
        yield hier
        nmisses = 0  # reset counter
        done.append(hier)


def make_type_hierarchy_subgraph(hierarchies: Iterable[Type[Hierarchy]]) -> Subgraph:
    subgraphs = []
    for hier in hierarchy_type_tree(hierarchies):
        labels = get_labels_of_schema_hierarchy(hier)
        node = Node(*labels, name=hier.__name__)
        bases = [Node(*get_labels_of_schema_hierarchy(b), name=b.__name__) for b in hier.__bases__ if b is not Hierarchy]
        rels = [Relationship(node, 'IS_TYPE_OF', b, order=i) for i, b in enumerate(bases)]
        subgraphs.append(Subgraph(bases + [node], rels))
    return reduce(lambda a, b: a | b, subgraphs)



def diff_hierarchy_schema_node(graph: Graph, hierarchy: Type[Hierarchy]):
    """
    Given a hierarchy, return py2neo subgraph that can be pushed to update the schema.
    If the new hierarchy is not backwards compatible, an exception will be raised

    A hierarchy can only be merged into the schema if existing data will still match the schema.
    i.e. if all below are true:
        has the same idname
        has the same factors or more
        has the same parents and children
            (additional parents and children can only be specified if they are optional)

    In the schema:
        (a)<-[:has_parent]-(b) indicates that (b) requires a parent (a) at instantiation time
        (a)-[:has_child]->(b) indicates that (a) requires a child (b) at instantiation time
    """
    # structure is [{labels}, is_optional, (minn, maxn), rel_idname, is_one2one, class_name]
    actual_parents = [(get_labels_of_schema_hierarchy(p, True), False, (1, 1), None, False, p.__name__) if not isinstance(p, Multiple)
                      else (get_labels_of_schema_hierarchy(p.node, True), p.minnumber == 0,
                            (p.minnumber, p.maxnumber), p.relation_idname, isinstance(p, One2One), p.node.__name__) for p in hierarchy.parents]
    actual_children = [(get_labels_of_schema_hierarchy(p, True), False, (1, 1), None, False, p.__name__) if not isinstance(p, Multiple)
                       else (get_labels_of_schema_hierarchy(p.node, True), p.minnumber == 0,
                             (p.minnumber, p.maxnumber), p.relation_idname, isinstance(p, One2One), p.node.__name__) for p in hierarchy.children]
    results = get_all_schema_rows(graph, hierarchy.__name__)
    labels = get_labels_of_schema_hierarchy(hierarchy)
    if len(results) == 0:  # node is completely new
        parents = []
        children = []
        found_node = Node(*labels,
                          factors=hierarchy.factors, idname=hierarchy.idname,
                          name=hierarchy.__name__,
                          singular_name=hierarchy.singular_name,
                          plural_name=hierarchy.plural_name,
                          identifier_builder=hierarchy.identifier_builder)
        for struct in actual_parents:
            props = dict(optional=struct[1], minnumber=struct[2][0], maxnumber=struct[2][1],
                         idname=struct[3], one2one=struct[4])
            parent = Node(*struct[0], name=struct[5])  # extant parent, specify labels so it matches not creates
            parents.append([parent, props])
        for struct in actual_children:
            props = dict(optional=struct[1], minnumber=struct[2][0], maxnumber=struct[2][1],
                         idname=struct[3], one2one=struct[4])
            child = Node(*struct[0], name=struct[5])  # extant child, specify labels so it matches not creates
            children.append([child, props])
    else:
        found_node = results[0][0]
        actual_parents = set(actual_parents)
        actual_children = set(actual_children)

        # gather extant info
        found_factors = set(found_node.get('factors', []))
        found_parents = {(frozenset(r[1].labels), bool(rel['optional']),
                          (int_or_none(rel['minnumber']), int_or_none(rel['maxnumber'])),
                          rel['idname'], bool(rel['one2one']), r[1]['name']) for r in results if r[1] is not None for rel in r[3]}
        found_children = {(frozenset(r[2].labels), bool(rel['optional']),
                           (int_or_none(rel['minnumber']), int_or_none(rel['maxnumber'])),
                           rel['idname'], bool(rel['one2one']), r[2]['name']) for r in results if r[2] is not None for rel in r[4]}

        # see if hierarchy is different in any way
        different_idname = found_node.get('idname') != hierarchy.idname
        different_identifier_builder = found_node.get('identifier_builder') != hierarchy.identifier_builder
        different_singular_name = found_node.get('singular_name') != hierarchy.singular_name
        different_plural_name = found_node.get('plural_name') != hierarchy.plural_name
        missing_factors = found_factors - set(hierarchy.factors)
        different_labels = set(labels).symmetric_difference(found_node.labels)

        # parents are different?
        missing_parents = found_parents - actual_parents
        new_parents = {p for p in actual_parents - found_parents if not p[1]}  # allowed if the new ones are optional

        # children are different?
        missing_children = found_children - actual_children  # missing from new definition
        new_children = {p for p in actual_children - found_children if not p[1]}  # allowed if the new ones are optional

        if different_idname or missing_factors or different_labels or missing_parents or \
                new_parents or missing_children or new_children or different_singular_name or different_plural_name:
            msg = f'Cannot add new hierarchy {hierarchy} because the {hierarchy.__name__} already exists' \
                  f' and the proposed definition of {hierarchy} is not backwards compatible. ' \
                  f'The differences are listed below:\n'
            if different_idname:
                msg += f'- proposed idname {hierarchy.idname} is different from the original {found_node.get("idname")}\n'
            if different_identifier_builder:
                msg += f'- proposed identifier_builder {hierarchy.identifier_builder} is different from the original {found_node.get("identifier_builder")}\n'
            if different_singular_name:
                msg += f'- proposed singular_name {hierarchy.singular_name} is different from the original {found_node.get("singular_name")}\n'
            if different_plural_name:
                msg += f'- proposed plural_name {hierarchy.plural_name} is different from the original {found_node.get("plural_name")}\n'
            if different_labels:
                msg += f'- proposed inherited types {different_labels} are different from {labels}\n'
            if missing_factors:
                msg += f'- factors {missing_factors} are missing from proposed definition\n'
            if new_parents:
                msg += f'- new parents with labels {[set(p[0]) - {"SchemaNode"} for p in new_parents]} are not optional (and therefore arent backwards compatible)\n'
            if new_children:
                msg += f'- new children with labels {[set(p[0]) - {"SchemaNode"} for p in new_children]} are not optional (and therefore arent backwards compatible)\n'
            if missing_parents:
                msg += f'- parents with labels {[set(p[0]) - {"SchemaNode"} for p in missing_parents]} are missing from the new definition\n'
            if missing_children:
                msg += f'- children with labels {[set(p[0]) - {"SchemaNode"} for p in missing_children]} are missing from the new definition\n'
            msg += f'any flagged children or parents may have inconsistent min/max number'
            raise AttemptedSchemaViolation(msg)

        nodes = []
        if set(hierarchy.factors) != found_factors:
            found_node['factors'] = hierarchy.factors  # update
            nodes.append(found_node)
        if found_children.symmetric_difference(actual_children):
            children = [(Node(*labels, name=name), dict(optional=optional, minnumber=minn, maxnumber=maxn,
                                             idname=idname, one2one=one2one))
                        for labels, optional, (minn, maxn), idname, one2one, name in actual_children]
            nodes += children
        else:
            children = []
        if found_parents.symmetric_difference(actual_parents):
            parents = [(Node(*labels, name=name), dict(optional=optional, minnumber=minn, maxnumber=maxn,
                                            idname=idname, one2one=one2one))
                       for labels, optional, (minn, maxn), idname, one2one, name in actual_parents]
            nodes += parents
        else:
            parents = []
    rels = []
    for c, props in children:
            rels.append(Relationship(found_node, 'HAS_CHILD', c, **props, originated=found_node.get('name')))
            if props['one2one']: # reflect the relationship if it is One2One
                rels.append(Relationship(c, 'HAS_CHILD', found_node, **props, originated=found_node.get('name')))
    for p, props in parents:
            rels.append(Relationship(p, 'IS_PARENT_OF', found_node, **props, originated=found_node.get('name')))
            if props['one2one']: # reflect the relationship if it is One2One
                rels.append(Relationship(found_node, 'IS_PARENT_OF', p, **props, originated=found_node.get('name')))
    return Subgraph([i[0] for i in parents + children] + [found_node], rels)


def write_schema(graph, hierarchies, dryrun=False, return_subgraph=False):
    """
    writes to the neo4j schema graph for use in optimising queries
    this should always be done before writing data
    """
    # sort available hierarchies meaning non-dependents first
    executions = []
    for hier in hierarchy_dependency_tree(hierarchies):
        subgraph = diff_hierarchy_schema_node(graph, hier)
        executions.append(push_py2neo_schema_subgraph_cypher(subgraph))
    executions.append(push_py2neo_schema_subgraph_cypher(make_type_hierarchy_subgraph(hierarchies)))
    if not dryrun:
        for cypher, params in tqdm(executions, desc='schema updates'):
            graph.execute(cypher, **params)
    return True


def get_dependencies_from_schema_rows(rows):
    parents = {row[1].get('name'): row[3][0] for row in rows if row[3]}
    children = {row[2].get('name'): row[4][0] for row in rows if row[4]}
    return parents, children


def parse_dependency(hierarchies: Dict[str, Type[Hierarchy]], dependency: Relationship) -> Union[Multiple, Type[Hierarchy]]:
    try:
        hierarchy = hierarchies[dependency.start_node['name']]
    except KeyError:
        try:
            hierarchy = hierarchies[dependency.end_node['name']]
        except KeyError:
            hierarchy = dependency.start_node['name']  # itself then
    mn, mx = dependency['minnumber'], dependency['maxnumber']
    if dependency['one2one']:
        dep = One2One(hierarchy, idname=dependency['name'])
    elif mn == 0 or mn is None:
        dep = Optional(hierarchy, idname=dependency['name'])
    elif mn == 1 and mx == 1 and dependency.get('name') is None:
        dep = hierarchy
    else:
        dep = Multiple(hierarchy, mn, mx, idname=dependency['name'])
    return dep


class SchemaNode(Hierarchy):
    idname='name'


def read_schema(graph) -> Set[Type[Hierarchy]]:
    bases_and_attrs = {r[0]: (r[1], r[2]) for r in get_all_schema_nodes(graph)}
    relation_tables = [(name, get_dependencies_from_schema_rows(get_all_schema_rows(graph, name))) for name, (bases, attrs) in bases_and_attrs.items()]
    relation_tables = deque(sorted(relation_tables, key=lambda x: len(x[1][0]) + len(x[1][1])))
    done = {'Hierarchy': Hierarchy, 'SchemaNode': SchemaNode}
    nmisses = 0
    while relation_tables:
        if nmisses > len(relation_tables): # a whole pass has been made without a success
            raise AttemptedSchemaViolation(f"Dependency resolution impossible, read schema elements may have cyclic dependencies:"
                                           f"{[i[0] for i in relation_tables]}")
        name, (parents, children) = relation_tables.popleft()
        base_names, attrs = bases_and_attrs[name]
        if not all(p in done for p in parents if p != name) or not all(c in done for c in children if c != name):
            logging.info(f"{name} put at the back because it requires dependencies which have not been written yet")
            relation_tables.append((name, (parents, children)))  # do it after the dependencies are done
            nmisses += 1
            continue
        try:
            bases = tuple([done[base] for base in base_names])
            if len(bases) == 0:
                bases = (Hierarchy,)
        except KeyError:
            nmisses += 1
            relation_tables.append((name, (parents, children)))  # do it after the dependencies are done
            continue
        parents = [parse_dependency(done, parent) for parent in parents.values()]
        children = [parse_dependency(done, child) for child in children.values()]
        attrs['parents'] = parents
        attrs['children'] = children
        if 'idname' not in attrs:
            attrs['is_template'] = True
            del attrs['parents']
            del attrs['children']
        done[name] = type(name, bases, attrs)
        nmisses = 0
    for k, v in done.items():
        v.instantate_nodes()
    return set(done.values()) - {SchemaNode, Hierarchy}


# TODO: put templates in and also label class dependency tree in neo4j



from collections import Counter
from contextlib import contextmanager
from typing import List, Callable, Union

from .base import CypherQuery, CypherVariable, Collection, CustomStatement, CypherAppendStr, Alias
from .statements import Unwind, Collect, GroupBy, Copy


def copy(*variables: CypherVariable):
    query = CypherQuery.get_context()  # type: CypherQuery
    g = Copy(*variables)
    query.add_statement(g)
    return g.output_variables


@contextmanager
def unwind(*args, enumerated=False, use_copy=True):
    if use_copy:
        args = copy(*args)
    query = CypherQuery.get_context()  # type: CypherQuery
    unwinder = Unwind(*args, enumerated=enumerated)
    query.open_context()  # allow this context to accumulate variables
    query.add_statement(unwinder)  # append the actual statement
    if len(unwinder.passed_outputs) == 1:
        yield unwinder.passed_outputs[0]  # give back the unwound variables
    else:
        yield tuple(unwinder.passed_outputs)
    query.close_context()  # remove the variables from being referenced from now on
    query.open_contexts = [[param for param in c if param not in args] for c in query.open_contexts]
    # previous = [param for context in query.open_contexts for param in context if param not in args]
    previous = [param for context in query.open_contexts for param in context]
    query.statements.append(Collect(previous))  # allow the collections to be accessible - force


def collect(*variables: CypherVariable, skip_nones=True):
    query = CypherQuery.get_context()  # type: CypherQuery
    collector = query.statements[-1]
    if not isinstance(collector, Collect):
        raise NameError(f"You must use collect straight after a with context")
    for variable in variables:
        if variable not in query.closed_context:
            if skip_nones and variable is None:
                continue
            raise ValueError(f"Cannot collect a non unwound variable {variable}")
    is_duplicated = [v in variables[:i] and v is not None for i, v in enumerate(variables)]  # True only for first instance
    to_collect = [v for dup, v in zip(is_duplicated, variables) if v is not None and not dup]
    collector = Collect(collector.previous, *to_collect)
    query.statements[-1] = collector
    r = [collector[v] if v is not None else None for v in variables]  # will pick out duplicates and None after the fact
    query.open_contexts[-1] += [i for i, dup in zip(r, is_duplicated) if i is not None and not dup]  # dont put duplicate variables into the stack
    if len(r) == 1:
        return r[0]
    return r


def groupby(variable_list, propertyname):
    if not isinstance(variable_list, Collection):
        raise TypeError(f"{variable_list} is not a collection")
    query = CypherQuery.get_context()  # type: CypherQuery
    g = GroupBy(variable_list, propertyname)
    query.add_statement(g)
    return g.output_variables[0]


def custom(statement: Callable, inputs: List[CypherVariable] = None, returns: List[Union[CypherVariable, str]] = None,
           outputs: List[Union[CypherVariable, str]] = None):
    if returns is None:
        returns = []
    if inputs is None:
        inputs = []
    if outputs is None:
        outputs = []
    for i, o in enumerate(returns):
        if isinstance(o, str):
            returns[i] = CypherVariable(o)
        elif not isinstance(o, CypherVariable):
            raise TypeError(f"output_variables  must be str or cyphervariable")
    for i, o in enumerate(outputs):
        if isinstance(o, str):
            outputs[i] = CypherVariable(o)
        elif not isinstance(o, CypherVariable):
            raise TypeError(f"output_variables  must be str or cyphervariable")
    outputs = returns + outputs
    statement = CustomStatement(statement, inputs, outputs)
    query = CypherQuery.get_context()  # type: CypherQuery
    query.add_statement(statement)
    return returns

def string_append(string, append, alias=False):
    appended = CypherAppendStr(string, append)
    if alias:
        query = CypherQuery.get_context()
        alias_statement = Alias(appended, 'append')
        query.add_statement(alias_statement)
        return alias_statement.out
    return appended


import operator
from collections.abc import Iterable, Callable
from typing import Any, overload

from typing_extensions import SupportsIndex, Literal

from .base import QueryFunctionBase
from math import floor, ceil



class QueryWrapper(QueryFunctionBase):

    def __init__(self, *unnamed_queries, **named_queries):
        self.unnamed_queries = unnamed_queries
        self.named_queries = named_queries
        self.queries = {**named_queries}
        self.queries.update({i: q for i, q in enumerate(unnamed_queries)})
        self.first_query = next(iter(self.queries.values()))

    def _precompile(self):
        return self

    def __repr__(self):
        return f'{self.__class__.__name__}({next(iter(self.queries.values()))} [{self.queries.keys()}])'

    def apply(self, function, *args, **kwargs):
        return apply(function, self, *args, **kwargs)

    def __getitem__(self, item):
        return self.apply(operator.__getitem__, item)

    def __getattr__(self, item):
        return self.apply(getattr, item)

    def __and__(self, other):
        return self.apply(operator.__and__, other)

    def __or__(self, other):
        return self.apply(operator.__or__, other)

    def __xor__(self, other):
        return self.apply(operator.__xor__, other)

    def __invert__(self):
        return self.apply(operator.__invert__)

    def __add__(self, other):
        return self.apply(operator.__add__, other)

    def __mul__(self, other):
        return self.apply(operator.__mul__, other)

    def __sub__(self, other):
        return self.apply(operator.__sub__, other)

    def __truediv__(self, other):
        return self.apply(operator.__truediv__, other)

    def __eq__(self, other):
        return self.apply(operator.__eq__, other)

    def __ne__(self, other):
        return self.apply(operator.__ne__, other)

    def __lt__(self, other):
        return self.apply(operator.__lt__, other)

    def __le__(self, other):
        return self.apply(operator.__le__, other)

    def __gt__(self, other):
        return self.apply(operator.__gt__, other)

    def __ge__(self, other):
        return self.apply(operator.__ge__, other)

    def __ceil__(self):
        return self.apply(ceil)

    def __floor__(self):
        return self.apply(floor)

    def __round__(self, ndigits: int):
        return self.apply(round, ndigits=ndigits)

    def __neg__(self):
        return self.apply(operator.__neg__)

    def __abs__(self):
        return self.apply(operator.__abs__)

    def __iadd__(self, other):
        return self.apply(operator.__iadd__, other)

    def __call__(self, *args, **kwargs):
        if isinstance(self.first_query, QueryFunctionBase):
            q = self._precompile()
            if q is not self:
                return q(*args, **kwargs)
            return {k: v(*args, **kwargs) for k, v in self.queries.items()}
        return self.apply(lambda x, *a, **k: x(*a, **k), *args, **kwargs)

    def __iter__(self):
        q = self._precompile()
        if isinstance(q, QueryWrapper):
            queries = {k: q._iterate() for k, q in self.queries.items()}
            while True:
                row = {k: next(v, None) for k, v in queries.items()}
                if all(v is None for v in row.values()):
                    break
                yield row
        else:
            yield from iter(q)

    def _debug_output(self, *args, **kwargs):
        return self._precompile()._debug_output(*args, **kwargs)

def crawl_copy_replace(obj, args_i=None, kwargs_k=None):
    """
    Return a shallow copy of the given object but with any QueryWrapper contained in it replaced with one of its subqueries denoted
    by args_i or kwargs_k
    The crawling only happens on the first level of the object and only for iterables
    """
    if isinstance(obj, QueryWrapper):
        if args_i is not None and kwargs_k is not None:
            raise ValueError('cannot specify both args_i and kwargs_k')
        elif args_i is not None:
            return obj.unnamed_queries[args_i]
        elif kwargs_k is not None:
            return obj.named_queries[kwargs_k]
        else:
            raise ValueError('Either args_i or kwargs_k must be specified')
    elif isinstance(obj, dict):
        return {crawl_copy_replace(k, args_i, kwargs_k): crawl_copy_replace(v, args_i, kwargs_k) for k, v in obj.items()}
    elif isinstance(obj, Iterable) and not isinstance(obj, str):
        return obj.__class__([crawl_copy_replace(v, args_i, kwargs_k) for v in obj])
    else:
        return obj

def find_query_wrapper(obj):
    """
    Return the first QueryWrapper found in the given object
    """
    if isinstance(obj, QueryWrapper):
        return obj
    elif isinstance(obj, dict):
        for v in obj.values():
            q = find_query_wrapper(v)
            if q is not None:
                return q
    elif isinstance(obj, Iterable):
        for v in obj:
            q = find_query_wrapper(v)
            if q is not None:
                return q
    return None


def apply(function, *args, **kwargs):
    """
    Apply a function to all queries contained within any arg or kwarg which is a QueryWrapper.
    All unnamed and named queries will be matched and fed to the function.
    The effect (with only one unnamed_query and one named query) is the same as:
    >>> function(*[a.unnamed_queries[0] for a in args], **{k: v.unnamed_queries[0] for k, v in kwargs.items()})
    But this function is more flexible and can be used to apply a function to all queries in a QueryWrapper
    and any number of args or kwargs can be QueryWrappers or any other type of object.
    """
    q = find_query_wrapper({'args': args, 'kwargs': kwargs})
    if q is None:
        return function(*args, **kwargs)
    unnamed_queries = [function(*crawl_copy_replace(args, args_i=i), **crawl_copy_replace(kwargs, args_i=i)) for i in range(len(q.unnamed_queries))]
    named_queries = {k: function(*crawl_copy_replace(args, kwargs_k=k), **crawl_copy_replace(kwargs, kwargs_k=k)) for k, v in q.named_queries.items()}

    # wrappers = [a for a in args if isinstance(a, QueryWrapper)] + [v for k, v in kwargs.items() if isinstance(v, QueryWrapper)]
    # n_unnamed_queries = len(wrappers[0].unnamed_queries) if wrappers else 0
    # n_named_queries = len(wrappers[0].named_queries) if wrappers else 0
    # if not all(len(w.unnamed_queries) == n_unnamed_queries for w in wrappers) and \
    #         not all(w.named_queries.keys() == wrappers[0].named_queries.keys() for w in wrappers):
    #     raise ValueError('All unnamed_queries and named_queries must be the same length. These QueryWrappers are not compatible.')
    # if not (n_unnamed_queries or n_named_queries):
    #     return function(*args, **kwargs)
    # unnamed_queries = [function(*[a.unnamed_queries[i] if isinstance(a, QueryWrapper) else a for a in args],
    #                             **{k: v.unnamed_queries[i] if isinstance(v, QueryWrapper) else v for k, v in kwargs.items()})
    #                    for i in range(n_unnamed_queries)]
    # named_queries = {k: function(*[a.named_queries[k] if isinstance(a, QueryWrapper) else a for a in args],
    #                                 **{k: v.named_queries[k] if isinstance(v, QueryWrapper) else v for k, v in kwargs.items()})
    #                     for k in wrappers[0].named_queries}
    return q.__class__(*unnamed_queries, **named_queries)


from textwrap import dedent
from typing import List, Dict, Union, Tuple, Optional, Iterable, Type

from . import CypherQuery
from .base import camelcase, Varname, Statement, CypherVariable, CypherData, CypherVariableItem, Collection


def are_different(a: str, b: str) -> str:
    return f"apoc.coll.different([apoc.coll.flatten([[{a}]]), apoc.coll.flatten([[{b}]])])"


def neo4j_dictionary(d: Union[dict, CypherVariable]) -> Tuple[Union[dict, CypherVariable], List[CypherVariable]]:
    """
    If d is a cyphervariable, return it
    If not, then we submit all the individual entries as data params to neo4j
    This avoids translating the data types ourselves!
    """
    if isinstance(d, CypherVariable):
        return d, [d]
    assert all(isinstance(k, str) for k in d.keys()), "keys must be strings"
    newd = {}
    ins = []
    for k, v in d.items():
        assert v is not None
        k = Varname(k)
        if not isinstance(v, CypherVariable):
            v = CypherData(v)
        ins.append(v)
        newd[k] = v
    return newd, ins


def sanitise_variablename(v):
    return f'`{str(v).replace("$", "")}`'


def expand_to_cypher_dict(*collections: Union[Dict[str, CypherVariable], CypherVariable]) -> str:
    inputs = []
    for collection in collections:
        if isinstance(collection, dict):
            inputs += list(collection.values())
        elif isinstance(collection, list):
            inputs += collection
        elif isinstance(collection, CypherVariable):
            inputs.append(collection)
        else:
            raise TypeError(f"Cannot convert {collection} to a cypher input dict of form `{{x:x}}` ")
    inputs = [getattr(i, 'parent') if isinstance(i, CypherVariableItem) else i for i in inputs]
    l = set([f"{sanitise_variablename(v)}: {v}" for v in inputs])
    return ', '.join(l)


def expand_to_cypher_alias(*collections: Union[Dict[str, CypherVariable], CypherVariable],
                           prefix='$') -> str:
    inputs = []
    for collection in collections:
        if isinstance(collection, dict):
            inputs += list(collection.values())
        elif isinstance(collection, list):
            inputs += collection
        elif isinstance(collection, CypherVariable):
            inputs.append(collection)
        else:
            raise TypeError(f"Cannot convert {collection} to a cypher input dict of form `{{x:x}}` ")
    inputs = [getattr(i, 'parent') if isinstance(i, CypherVariableItem) else i for i in inputs]
    l = set([f"{prefix}{sanitise_variablename(v)} as {sanitise_variablename(v)}" for v in inputs])
    return ', '.join(l)


class MatchNode(Statement):
    keyword = 'MATCH'

    def __init__(self, labels: List[str], properties: dict, optional=False):
        self.labels = [camelcase(l) for l in labels]
        self.properties, inputs = neo4j_dictionary(properties)
        self.out = CypherVariable(labels[0])
        self.optional = optional
        if optional:
            self.keyword = 'OPTIONAL MATCH'
        super(MatchNode, self).__init__(inputs, [self.out])

    def to_cypher(self):
        labels = ':'.join(map(str, self.labels))
        return f"{self.keyword} ({self.out}:{labels} {self.properties})"


class MatchRelationship(Statement):
    keyword = 'MATCH'

    def __init__(self, parent, child, reltype: str, properties: dict, optional=False):
        self.parent = parent
        self.child = child
        self.reltype = reltype
        self.properties, inputs = neo4j_dictionary(properties)
        self.optional = optional
        if optional:
            self.keyword = 'OPTIONAL MATCH'
        inputs += [self.parent, self.child]
        self.out = CypherVariable(reltype)
        super().__init__(inputs, [self.out])

    def to_cypher(self):
        if self.reltype is None and not len(self.properties):
            reldata = f'[{self.out}]'
        elif self.reltype is None:
            reldata = f'[{self.out}: {self.properties}]'
        else:
            reldata = f'[{self.out}:{self.reltype} {self.properties}]'
        return f"{self.keyword} ({self.parent})-{reldata}->({self.child})"


class MatchIdNode(Statement):
    def __init__(self, labels: List[str], id: Union[int, CypherVariable]):
        self.labels = [camelcase(l) for l in labels]
        if isinstance(id, CypherVariable):
            self.id = id
        else:
            self.id = CypherData(id)
        self.out = CypherVariable(labels[0])
        super().__init__([self.id], [self.out], [])

    def to_cypher(self):
        labels = ':'.join(self.labels)
        return f'MATCH ({self.out}: {labels}) where id({self.out}) = {self.id}'


class MatchPatternNode(Statement):
    def __init__(self, labels: List[str], properties: Dict[str, Union[str, int, float, CypherVariable]],
                 parents: List[CypherVariable], children: List[CypherVariable], exclude: List[CypherVariable]):
        self.labels = [camelcase(l) for l in labels]
        self.properties, inputs = neo4j_dictionary(properties)
        self.parents = [p for p in parents if not isinstance(p, Collection)]
        self.collection_parents = [p for p in parents if isinstance(p, Collection)]
        self.children = [c for c in children if not isinstance(c, Collection)]
        self.collection_children = [c for c in children if isinstance(c, Collection)]
        self.exclude = [e for e in exclude if not isinstance(e, Collection)]
        self.exclude_collection = [e for e in exclude if isinstance(e, Collection)]
        self.out = CypherVariable(labels[0])
        super().__init__(inputs+parents+children+exclude, [self.out], [])

    def to_cypher(self):
        labels = ':'.join(self.labels)
        match = f'({self.out}: {labels} {self.properties})'
        relations = ', '.join([f'({self.out})<--({p})' for p in self.parents] +
                                [f'({self.out})-->({c})' for c in self.children])
        extras = ' AND '.join([f"({self.out} <> {e} or {e} is null)" for e in self.exclude] +
                              [f"all(n in {e} where {self.out} <> n or n is null)" for e in self.exclude_collection] +
                              [f"all(p in {p} where exists( ({self.out})<--(p) ))" for p in self.collection_parents] +
                              [f"all(c in {c} where exists( ({self.out})-->(c) ))" for c in self.collection_children])
        s = f'WITH * OPTIONAL MATCH {match}'
        if relations:
            s += f',{relations}'
        if extras:
            s += f' WHERE {extras}'
        return s

class MatchBranchNode(Statement):
    def __init__(self, *nodes_or_labels):
        self.nodes_or_labels = nodes_or_labels
        self.out = CypherVariable(nodes_or_labels)
        inputs = [i for i in nodes_or_labels if isinstance(i, CypherVariable)]
        outputs = [i  if isinstance(i, CypherVariable) else CypherVariable(i) for i in nodes_or_labels]
        super().__init__(inputs, outputs, [])

    def to_cypher(self):
        nodes = [f'({o}:{n})' if isinstance(n, str) else f'({o})' for n, o in zip(self.nodes_or_labels, self.output_variables)]
        path = '<-[:is_required_by]-'.join(map('{}'.format, nodes))
        return f'WITH * OPTIONAL MATCH {path}'

class PropertyOverlapError(Exception):
    pass


class NullPropertyError(Exception):
    pass


class CollisionManager(Statement):
    def __init__(self, out, identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]], collision_manager='track&flag'):
        self.out = out
        self.properties, propinputs = neo4j_dictionary(properties)
        self.identproperties, identinputs = neo4j_dictionary(identproperties)
        self.validate_properties()
        self.propvar = CypherVariable('props')
        self.colliding_keys = CypherVariable('colliding_keys')
        self.value = CypherVariable('unnamed')
        inputs = propinputs + identinputs
        outputs = [self.out, self.propvar]
        if collision_manager == 'track&flag':
            outputs += [self.value, self.colliding_keys]
        elif collision_manager not in ['overwrite', 'ignore']:
            raise ValueError(f"Unknown collision_manager {collision_manager}")
        self.collision_manager = collision_manager
        super().__init__(inputs, outputs)

    def validate_properties(self):
        if any(p in self.identproperties for p in self.properties.keys()):
            raise PropertyOverlapError(f"Cannot have the same key in both properties and identproperties")
        for k, v in self.identproperties.items():
            if v != v:
                raise NullPropertyError(f"Cannot assign a nan to a node identify property")

    @property
    def on_match(self):
        if self.collision_manager == 'overwrite':
            return f"SET {self.out} += {self.propvar}   // overwrite with new colliding properties"
        return f"SET {self.out} = apoc.map.merge({self.propvar}, properties({self.out}))   // update, keeping the old colliding properties"

    @property
    def on_create(self):
        return f"SET {self.out}._dbcreated = time0, {self.out} += {self.propvar}  // setup as standard"

    @property
    def on_run(self):
        return f'SET {self.out}._dbupdated = time0  // always set updated time '

    @property
    def merge_statement(self):
        raise NotImplementedError

    @property
    def collision_record(self):
        raise NotImplementedError

    @property
    def collision_record_input(self):
        raise NotImplementedError

    @property
    def post_merge(self):
        return dedent(f"""
    ON MATCH {self.on_match}
    ON CREATE {self.on_create}
    {self.on_run}""")

    @property
    def pre_merge(self):
        return f"WITH *, {self.properties} as {self.propvar}"

    @property
    def merge_paragraph(self):
        return f"""
        {self.pre_merge}
        {self.merge_statement}
        {self.post_merge}
        """

    def to_cypher(self):
        query = self.merge_paragraph
        if self.collision_manager == 'track&flag':
            query += f"""
            WITH *, [x in apoc.coll.intersection(keys({self.propvar}), keys(properties({self.out}))) where ({self.propvar}[x] is null or {self.out}[x] is null) or {self.propvar}[x] <> {self.out}[x]] as {self.colliding_keys}
            CALL apoc.do.when(size({self.colliding_keys}) > 0, 
                "{self.collision_record} SET c = $collisions SET c._dbcreated = $time RETURN $time", 
                "RETURN $time",
                {{{self.collision_record_input}, collisions: apoc.map.fromLists({self.colliding_keys}, apoc.map.values({self.propvar}, {self.colliding_keys})), time:time0}}) yield value as {self.value}
            """
        return dedent(query)


class MergeNode(CollisionManager):
    def __init__(self, labels: List[str], identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]], collision_manager='track&flag'):
        self.labels = [camelcase(l) for l in labels]
        out = CypherVariable(labels[0])
        super().__init__(out, identproperties, properties, collision_manager)

    @property
    def merge_statement(self):
        labels = ':'.join(map(str, self.labels))
        return f'MERGE ({self.out}: {labels} {self.identproperties})'

    @property
    def collision_record(self):
        return f"WITH $innode as innode CREATE (c:_Collision)-[:COLLIDES]->(innode)"

    @property
    def collision_record_input(self):
        return f"innode: {self.out}"


class MergeRelationship(CollisionManager):
    def __init__(self, parent, child, reltype: str, identproperties: dict, properties: dict, collision_manager='track&flag'):
        self.parent = parent
        self.child = child
        self.reltype = reltype
        out = CypherVariable(reltype)
        self.value = CypherVariable('value')
        super().__init__(out, identproperties, properties, collision_manager)
        self.output_variables.append(self.value)

    def to_cypher(self):
        return super().to_cypher()

    @property
    def merge_statement(self):
        return f'call apoc.merge.relationship($parent, "{self.reltype}", $ident, $props, $child, $onmatch) yield rel as {self.out}'


    @property
    def merge_paragraph(self):
        return f"""{self.pre_merge} 
        call apoc.do.when({self.parent} is null or {self.child} is null, 'return null as {self.out}','
        {self.merge_statement}
        RETURN {self.out}', {{parent: {self.parent}, child:{self.child}, ident:{self.identproperties}, 
                              props: {self.propvar}, onmatch: {{}}}}) yield value as {self.value}
        OPTIONAL MATCH ({self.parent})-[{self.out}:{self.reltype} {self.identproperties}]->({self.child})
        """

    @property
    def collision_record(self):
        return f"WITH $a as a, $b as b CREATE (a)-[c:COLLIDES]->(b)"

    @property
    def collision_record_input(self):
        return f"a:{self.parent}, b:{self.child}"


class MergeDependentNode(CollisionManager):
    def __init__(self, labels: List[str], identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]],
                 parents: List[CypherVariable], anti_id_rels: Tuple[str, bool, Dict],
                 reltypes: List[str],
                 reldirs: List[bool], # True if parent, False if child
                 relidentproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                 relproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                 collision_manager='track&flag'):
        if not (len(parents) == len(reltypes) == len(relproperties) == len(relidentproperties)):
            raise ValueError(f"Parents must have the same length as reltypes, relproperties, relidentproperties")
        self.labels = [camelcase(l) for l in labels]
        self.relidentproperties, relidentpropins = [], []
        self.relproperties, relpropins = [], []
        for ident, prop in zip(relidentproperties, relproperties):
            identdict, identpropins = neo4j_dictionary(ident)
            propdict, propins = neo4j_dictionary(prop)
            self.relidentproperties.append(identdict)
            self.relproperties.append(propdict)
            relidentpropins += identpropins
            relpropins += propins
        self.reldirs = reldirs
        self.parents = parents
        self.outnode = CypherVariable(labels[0])
        self.relvars = [CypherVariable(reltype) for reltype in reltypes]
        self.dummyrelvars = [CypherVariable('dummy'+reltype) for reltype in reltypes]
        self.dummy = CypherVariable('dummy')
        self.reltypes = reltypes
        self.relpropsvars = [CypherVariable(f'{t}_props') for t in reltypes]
        self.colliding_rel_keys = [CypherVariable('colliding_rel_keys') for _ in reltypes]
        _anti_rel_ids = []
        anti_id_rel_inputs = []
        for label, reltype, reldir, props in anti_id_rels:
            d, anti_id_rel_input = neo4j_dictionary(props)
            anti_id_rel_inputs += anti_id_rel_input
            _anti_rel_ids.append((label, reltype, reldir, d))
        self.anti_id_rels = _anti_rel_ids
        super().__init__(self.outnode, identproperties, properties, collision_manager)
        self.child_holder = CypherVariable('child_holder')
        self.unnamed = CypherVariable('unnamed')
        self.hashvar = CypherVariable('hash')
        self.input_variables += parents
        self.input_variables += relidentpropins
        self.input_variables += relpropins
        self.input_variables += anti_id_rel_inputs
        self.output_variables += self.relvars
        self.output_variables += self.dummyrelvars
        self.output_variables += self.relpropsvars
        self.hidden_variables += self.colliding_rel_keys
        self.output_variables.append(self.dummy)
        self.output_variables.append(self.child_holder)
        self.output_variables.append(self.unnamed)
        self.hidden_variables.append(self.hashvar)

    def validate_properties(self):
        super(MergeDependentNode, self).validate_properties()
        for idents, props in zip(self.relidentproperties, self.relproperties):
            if any(p in idents for p in props.keys()):
                raise ValueError(f"Cannot have the same key in both properties and identproperties")
            for k, v in idents.items():
                if v != v:
                    raise NullPropertyError(f"Cannot assign a nan to a node identify property")

    @property
    def pre_merge(self):
        line = f"WITH *, {self.properties} as {self.propvar}"
        for relprop, relpropsvar in zip(self.relproperties, self.relpropsvars):
            line += f', {relprop} as {relpropsvar}'
        return line

    @property
    def merge_statement(self):
        labels = ':'.join(map(str, self.labels))
        anti_relations = []
        real_relations = []
        temp_relations = []
        test_relations = []
        parent_list = []
        for label, reltype, reldir, props in self.anti_id_rels:
            arrows = ['-' if reldir else '<-', '->' if reldir else '-']
            rel = f'(:{label}){arrows[0]}[:{reltype} {props}]{arrows[1]}({self.dummy})'
            anti_relations.append(rel)
        anti_relations = ' OR '.join(anti_relations)
        for i, (parent, reltype, reldir, relidentprop, dummyrelvar, relvar) in enumerate(zip(self.parents, self.reltypes, self.reldirs, self.relidentproperties, self.dummyrelvars, self.relvars)):
            child = f'({self.dummy})'
            arrows = ['-' if reldir else '<-', '->' if reldir else '-']
            rel = f'({parent}){arrows[0]}[{dummyrelvar}:{reltype} {relidentprop}]{arrows[1]}'
            relidentprop[Varname('_query_hash')] = self.hashvar
            real_rel = f'({parent}){arrows[0]}[{relvar}:{reltype} {relidentprop}]{arrows[1]}'
            if i == 0:
                first_child = f'({self.dummy}: {labels} {self.identproperties})'
                real_child = f'({self.out}: {labels} {self.identproperties})'
            else:
                real_child = f'({self.out})'
            real_relations.append(real_rel + real_child)
            temp_relations.append(rel + '(temp)')
            test_relations.append(rel + child)
            parent_list.append(f"{parent}")
        r = [{k: v for k, v in r.items() if k != Varname('_query_hash')} for r in self.relidentproperties]
        dct = expand_to_cypher_dict(self.dummy, self.propvar, self.identproperties, *self.parents + r + self.dummyrelvars)
        aliases = expand_to_cypher_alias(self.identproperties, *self.parents + r)
        variables = set()
        for p in self.parents:
            variables.add(p)
        for v in self.identproperties.values():
            try:
                v = v.parent
            except AttributeError:
                pass
            if not isinstance(v, CypherData):
                variables.add(v)
        for rel in self.relidentproperties:
            for v in rel.values():
                try:
                    v = v.parent
                except AttributeError:
                    pass
                if not isinstance(v, CypherData):
                    variables.add(v)
        hashes = [f"apoc.hashing.fingerprinting([{i}, '{rt}', {ri}], {{strategy: 'EAGER'}})" for i, (rt, ri) in enumerate(zip(self.reltypes, r))]
        hashes.append(f"apoc.hashing.fingerprinting({self.identproperties}, {{strategy: 'EAGER'}})")
        hashes = "with *, apoc.hashing.fingerprinting([" + ",  ".join(hashes) + f"], {{strategy: 'EAGER'}}) as {self.hashvar}"

        # merge_real_relations = '\n'.join([f'MERGE ({self.out}: {labels} {self.identproperties})'] + [f'MERGE {r}' for r in real_relations])
        merge_real_relations = '\n'.join([f'MERGE {r}' for r in real_relations])
        on_create_rel_returns = ', '.join([f'{relvar}' for relvar in self.relvars])
        on_match_rel_returns = ', '.join([f'${dummy} as {real}' for dummy, real in zip(self.dummyrelvars, self.relvars)])
        rel_expansion = expand_to_cypher_alias(self.out, *self.relvars, prefix=f'{self.child_holder}.')
        exists = ','.join([f'{tr}' for tr in test_relations])
        notexists_condition = f'WHERE NOT ({anti_relations})' if anti_relations else ''
        already_exists = f"OPTIONAL MATCH {first_child},{exists} {notexists_condition}"
        condition = f"{self.dummy} is null"
        iftrue = f"""
        WITH {aliases}, $time0 as time0
        {hashes}
        {merge_real_relations}
        SET {self.out} += ${self.propvar}
        RETURN {self.out}, {on_create_rel_returns}
        """
        iffalse = f"RETURN ${self.dummy} as {self.out}, {on_match_rel_returns}"
        when = f'CALL apoc.do.when({condition}, "{iftrue}", "{iffalse}", {{ {dct}, time0:time0}}) yield value as {self.child_holder}'
        when += f"\n WITH *, {rel_expansion}"
        return dedent(f"CALL apoc.lock.nodes({self.parents})\n{already_exists}\n{when}")

    @property
    def on_match(self):  # remember, we are in a call context
        query = ''
        for i, (r, rprops, colliding_keys) in enumerate(zip(self.relvars+[self.out], self.relpropsvars+[self.propvar], self.colliding_rel_keys+[self.colliding_keys])):
            if self.collision_manager == 'track&flag':
                raise NotImplementedError(f"This feature has not been adapted yet")
                if r != self.out:  # handled by the base class above
                    query += dedent(f"""
                        WITH *, [x in apoc.coll.intersection(keys({rprops}), keys(properties({r}))) where ({rprops}[x] is null or {r}[x] is null) or {rprops}[x] <> {r}[x]] as {colliding_keys}
                        CALL apoc.do.when(size({colliding_keys}) > 0, 
                            'WITH $inrel as inrel 
                             MATCH (a)-[inrel]->(b)  
                             CREATE (a)-[c:_Collision]->(b) SET c = $collisions 
                             SET c._dbcreated = $time
                             SET c._reltype = type(inrel)
                             RETURN $time', 
                            'RETURN $time',
                            {{inrel: {r}, collisions: apoc.map.fromLists({colliding_keys}, apoc.map.values({rprops}, {colliding_keys})), time:$time0}}) yield value as _{i}{self.value}
                    """)
            elif self.collision_manager == 'overwrite':
                query += f"\nSET {r} += {rprops}   // overwrite with new colliding properties"
            else:
                query += f"\nSET {r} = apoc.map.merge({rprops}, properties({r}))  // update, keeping the old colliding properties"
        return query

    @property
    def on_create(self):  # remember, we are in a call context
        query = f"SET {self.out}._dbcreated = $time0, {self.out} += {self.propvar}  // setup as standard"
        for r, rprops in zip(self.relvars, self.relpropsvars):
            query += f'\nSET {r}._dbupdated = $time0, {r}._dbcreated = $time0, {r} += {rprops}'
        return query

    @property
    def on_run(self):  # remember, we are in a call context
        query = f"SET {self.out}._dbupdated = time0  // always set updated time"
        for r in self.relvars:
            query += f'\nSET {r}._dbupdated = time0'
        return query

    @property
    def post_merge(self):
        dct = expand_to_cypher_dict(self.out, self.propvar, *self.relvars+self.relpropsvars)
        if len(dct):
            dct += ', '
        dct += 'time0:time0'
        aliases = expand_to_cypher_alias(self.out, self.propvar, *self.relvars+self.relpropsvars)
        return dedent(f"""
        // post merge
        call apoc.do.when({self.dummy} is null,
        "WITH {aliases}\n{self.on_create}\n RETURN $time0",
        "WITH {aliases}\n{self.on_match}\n RETURN $time0",
        {{ {dct} }}) yield value as {self.unnamed}
        {self.on_run}\n""")

    @property
    def collision_record(self):
        return f"WITH $innode as innode CREATE (c:_Collision)-[:COLLIDES]->(innode)"

    @property
    def collision_record_input(self):
        return f"innode: {self.out}"

    def to_cypher(self):
        return super().to_cypher()


class SetVersion(Statement):
    def __init__(self, parents: List[CypherVariable], reltypes: List[str], childlabel: str, child: CypherVariable, childproperties: dict):
        if len(reltypes) != len(parents):
            raise ValueError(f"reltypes must be the same length as parents")
        self.parents = parents
        self.reltypes = reltypes
        self.childlabel = camelcase(childlabel)
        self.childproperties, other_ins = neo4j_dictionary(childproperties)
        self.child = child
        other_ins.append(child)
        super(SetVersion, self).__init__(self.parents+other_ins, [])

    def to_cypher(self):
        matches = ', '.join([f'({p})-[:{r}]->(c:{self.childlabel} {self.childproperties})' for p, r in zip(self.parents, self.reltypes)])
        query = [
            f"WITH * CALL {{",
                f"\t WITH {','.join(map(str, self.parents))}, {self.child}",
                f"\t OPTIONAL MATCH {matches}"
                f"\t WHERE id(c) <> id({self.child})",
                f"\t WITH {self.child}, max(c.version) as maxversion",
                f"\t SET {self.child}.version = coalesce({self.child}['version'], maxversion + 1, 0)",
                f"\t RETURN {self.child}['version']",
            f"}}"
        ]
        return '\n'.join(query)


class ValidateNumber(Statement):
    def __init__(self, collection: Collection, mn: int, mx: int, central_typelabel: str, to_join_typelabel: str):
        self.collection = collection
        self.mn = mn
        self.mx = mx
        self.central_typelabel = central_typelabel
        self.to_join_typelabel = to_join_typelabel
        self.dummy = CypherVariable('check')
        super(ValidateNumber, self).__init__([self.collection], [], [self.dummy])

    def to_cypher(self):
        return f"WITH *, apoc.util.validatePredicate((size({self.collection}) > {self.mx}) or (size({self.collection}) < {self.mn}), " \
               f"'{self.central_typelabel} expects {self.mn} <= count({self.to_join_typelabel}) <= {self.mx} not %d', [size({self.collection})]) AS {self.dummy}"



class ValidateType(Statement):
    def __init__(self, variable: CypherVariable, typelabel: str):
        self.variable = variable
        self.typelabel = typelabel
        self.dummy = CypherVariable('check')
        hidden = [self.dummy]
        if isinstance(self.variable, Collection):
            self.x = CypherVariable('i')
            hidden.append(self.x)
        super(ValidateType, self).__init__([self.variable], [], hidden)

    def to_cypher(self):
        if isinstance(self.variable, Collection):
            return f"WITH *, apoc.util.validatePredicate(size([{self.x} in {self.variable} where NOT '{self.typelabel}' in labels({self.x})]) > 0, " \
                   f"'{self.variable} needs to be a collection of {self.typelabel}', []) as {self.dummy}"
        return f"WITH *, apoc.util.validatePredicate('NOT {self.typelabel}' in labels({self.variable}), '{self.variable} expects {self.typelabel}', []) as {self.dummy}"


def match_node(labels, properties, optional=False):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchNode(labels, properties, optional)
    query.add_statement(statement)
    return statement.out


def match_relationship(parent, child, reltype, properties, optional=False):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchRelationship(parent, child, reltype, properties, optional)
    query.add_statement(statement)
    return statement.out


def match_pattern_node(labels: List[str], properties: Dict[str, Union[str, int, float, CypherVariable]] = None,
                       parents: List[CypherVariable] = None, children: List[CypherVariable] = None,
                       exclude: List[CypherVariable] = None):
    query = CypherQuery.get_context()  # type: CypherQuery
    if properties is None:
        properties = {}
    if parents is None:
        parents = []
    if children is None:
        children = []
    if exclude is None:
        exclude = []
    statement = MatchPatternNode(labels, properties, parents, children, exclude)
    query.add_statement(statement)
    return statement.out

def match_id_node(labels: List[str], id: int):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchIdNode(labels, id)
    query.add_statement(statement)
    return statement.out


def match_branch_node(*nodes_or_labels: Union[str, Type[CypherVariable]]):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchBranchNode(*nodes_or_labels)
    query.add_statement(statement)
    return statement.output_variables


def merge_single_node(labels, identproperties, properties, collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeNode(labels, identproperties, properties, collision_manager)
    query.add_statement(statement)
    return statement.out


def merge_relationship(parent, child, reltype, identproperties, properties, collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeRelationship(parent, child, reltype, identproperties, properties, collision_manager)
    query.add_statement(statement)
    return statement.out


def merge_dependent_node(labels, identproperties, properties, parents, anti_id_rels,
                         reltypes, reldirs, relidentproperties, relproperties,
                         collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeDependentNode(labels, identproperties, properties, parents, anti_id_rels, reltypes, reldirs,
                                   relidentproperties, relproperties, collision_manager)
    query.add_statement(statement)
    return statement.outnode


def set_version(parents, reltypes, childlabel, child, childproperties):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = SetVersion(parents, reltypes, childlabel, child, childproperties)
    query.add_statement(statement)


def validate_number(collection, mn, mx, central_typelabel, to_join_typelabel):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = ValidateNumber(collection, mn, mx, central_typelabel, to_join_typelabel)
    query.add_statement(statement)

def validate_type(collection, typelabel):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = ValidateType(collection, typelabel)
    query.add_statement(statement)


def merge_node(labels, identproperties, properties=None,
               id_rels: Dict[CypherVariable, Union[Tuple[str, bool, Optional[Dict], Optional[Dict]], str]] = None,
               anti_id_rels: List[Tuple[str, bool, Dict]] = None,
               versioned_label=None,
               versioned_properties=None,
               collision_manager='track&flag') -> CypherVariable:
    if properties is None:
        properties = {}
    if id_rels is None:
        id_rels = {}
    if anti_id_rels is None:
        anti_id_rels = {}
    rel_list = []
    reldir_list = []
    reltype_list = []
    relidentproperties_list = []
    relproperties_list = []
    for rel, reldata in id_rels.items():
        if isinstance(reldata, str):
            reldata = [True, *reldata]
        rel_list.append(rel)
        reltype_list.append(reldata[0])
        reldir_list.append(reldata[1])
        if len(reldata) > 1:
            relidentproperties_list.append(reldata[2])
        else:
            relidentproperties_list.append({})
        if len(reldata) > 2:
            relproperties_list.append(reldata[3])
        else:
            relproperties_list.append({})
    if len(id_rels) or len(anti_id_rels):
        node = merge_dependent_node(labels, identproperties, properties, rel_list, anti_id_rels,
                                    reltype_list, reldir_list, relidentproperties_list, relproperties_list,
                                    collision_manager)
    else:
        node = merge_single_node(labels, identproperties, properties, collision_manager)
    if versioned_label is not None:
        if versioned_properties is None:
            versioned_properties = {}
        set_version(rel_list, reltype_list, versioned_label, node, versioned_properties)
    return node



from functools import partial
from typing import Union, Callable

import numpy as np
import py2neo
from astropy.table import Table

from .common import FrozenQuery, is_regex
from .tree import Branch
from ..hierarchy import Hierarchy
from ..writequery import CypherVariable, CypherQuery

__all__ = ['sum', 'min', 'max', 'std', 'all', 'any', 'count', 'abs', 'floor', 'ceil', 'round',
           'random', 'sign', 'log', 'log10', 'exp', 'sqrt', 'string', 'lower', 'upper']


def _template_aggregator(string_function, name, normal: Callable, item: 'Dissociated', wrt: FrozenQuery = None,
                         remove_infs: bool = True, convert_to_float: bool = False, allow_hiers=False,
                         args=None, kwargs=None):
    if not isinstance(item, FrozenQuery):
        if wrt is not None:
            args = (wrt, ) + args
        return normal(item, *args, **kwargs)
    elif wrt is None:
        branch = item.branch.handler.entry
    else:
        branch = wrt.branch
    if isinstance(item, Dissociated):
        itembranch, itemvariable = _convert(item.branch, item.variable, remove_infs, convert_to_float)
        new = branch.aggregate(string_function, itemvariable, itembranch, namehint=name)
    elif allow_hiers:
        new = branch.aggregate(string_function, item.hierarchy_variable, item.branch, namehint=name)
    else:
        raise TypeError(f"Calling {name} with {item} of type {type(item)} is not supported")
    return Dissociated(item.handler, new, new.action.target)


def _convert(branch: Branch, variable: CypherVariable, remove_infs: bool = True, convert_to_float: bool = False):
    if convert_to_float:
        branch = branch.operate('CASE WHEN {z} = true THEN 1 WHEN {z} = false THEN 0 ELSE {z} END', namehint='convert_to_float', z=variable)
        variable = branch.action.output_variables[0]
    if remove_infs:
        branch = branch.operate("CASE WHEN {variable} > apoc.math.maxLong() THEN null ELSE {variable} END", namehint='remove_infs', variable=variable)
        variable = branch.action.output_variables[0]
    return branch, variable


def _template_operator(string_function, name, normal: Callable, item: 'Dissociated', remove_infs: bool = True, convert_to_float: bool = False, args=None, kwargs=None):
    if not isinstance(item, FrozenQuery):
        return normal(item, *args, **kwargs)
    branch, variable = _convert(item.branch, item.variable, remove_infs, convert_to_float)
    branch = branch.operate(string_function, x=variable, namehint=name)
    return Dissociated(item.handler, branch, branch.action.output_variables[0])


python_abs = abs
python_round = round


def abs(item, *args, **kwargs):
    return _template_operator('abs({x})', 'abs', python_abs, item, convert_to_float=True, args=args, kwargs=kwargs)


def ceil(item, *args, **kwargs):
    return _template_operator('ceil({x})', 'ceil', np.ceil, item, convert_to_float=True, args=args, kwargs=kwargs)


def floor(item, *args, **kwargs):
    return _template_operator('floor({x})', 'floor', np.floor, item, convert_to_float=True, args=args, kwargs=kwargs)


def random(item, *args, **kwargs):
    return _template_operator('random({x})', 'random', np.uniform, item, convert_to_float=True, args=args, kwargs=kwargs)


def round(item, precision=0, *args, **kwargs):
    return _template_operator('round({x})', 'round', partial(python_round, precision), item, convert_to_float=True, args=args, kwargs=kwargs)


def sign(item, *args, **kwargs):
    return _template_operator('sign({x})', 'sign', np.sign, item, convert_to_float=True, args=args, kwargs=kwargs)


def exp(item, *args, **kwargs):
    return _template_operator('exp({x})', 'exp', np.exp, item, convert_to_float=True, args=args, kwargs=kwargs)


def log(item, *args, **kwargs):
    return _template_operator('log({x})', 'log', np.log, item, convert_to_float=True, args=args, kwargs=kwargs)


def log10(item, *args, **kwargs):
    return _template_operator('log10({x})', 'log10', np.log10, item, convert_to_float=True, args=args, kwargs=kwargs)


def sqrt(item, *args, **kwargs):
    return _template_operator('sqrt({x})', 'sqrt', np.sqrt, item, convert_to_float=True, args=args, kwargs=kwargs)


def string(item):
    if isinstance(item, Hierarchy):
        item = item.identifier
    elif hasattr(item, 'hierarchy_type'):
        item = getattr(item, item.hierarchy_type.idname)
    return _template_operator('toString({x})', 'string', str, item, convert_to_float=False)


def lower(item):
    item = string(item)
    return _template_operator('toLower({x})', 'lower', lambda x: x.lower(), item, convert_to_float=False)


def upper(item):
    item = string(item)
    return _template_operator('toUpper({x})', 'upper', lambda x: x.upper(), item, convert_to_float=False)



python_any = any
python_all = all
python_max = max
python_min = min
python_sum = sum


def sum(item, wrt=None, *args, **kwargs):
    return _template_aggregator('sum({x})', 'sum', python_sum, item, wrt, convert_to_float=True, args=args, kwargs=kwargs)


def max(item, wrt=None, *args, **kwargs):
    return _template_aggregator('max({x})', 'max', python_max, item, wrt, convert_to_float=True, args=args, kwargs=kwargs)


def min(item, wrt=None, *args, **kwargs):
    return _template_aggregator('min({x})', 'min', python_min, item, wrt, convert_to_float=True, args=args, kwargs=kwargs)


def all(item, wrt=None, *args, **kwargs):
    return _template_aggregator('all(i in collect({x}) where i)', 'all', python_all, item, wrt, args=args, kwargs=kwargs)


def any(item, wrt=None, *args, **kwargs):
    return _template_aggregator('any(i in collect({x}) where i)', 'any', python_any, item, wrt, args=args, kwargs=kwargs)


def count(item, wrt=None, *args, **kwargs):
    return _template_aggregator('count({x})', 'count', len, item, wrt, allow_hiers=True, args=args, kwargs=kwargs)


def std(item, wrt=None, *args, **kwargs):
    return _template_aggregator('stDev({x})', 'std', np.std, item, wrt, convert_to_float=True, args=args, kwargs=kwargs)


class Dissociated(FrozenQuery):
    def __init__(self, handler, branch: Branch, variable: CypherVariable, parent: 'FrozenQuery' = None):
        super().__init__(handler, branch, parent)
        self.variable = variable

    def _filter_by_boolean(self, boolean_filter: 'FrozenQuery'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.variable, self)

    def __getitem__(self, boolean_filter: 'Dissociated'):
        if not isinstance(boolean_filter, Dissociated):
            raise TypeError(f"Factors may only be filtered with `[]` using a boolean filter")
        return self._filter_by_boolean(boolean_filter)

    def _apply_aligning_func(self, string, other: 'Dissociated' = None):
        if other is not None:
            if isinstance(other, Dissociated):  # now we have to align first
                aligned = self.branch.align(other.branch)
                inputs = {
                    'x': aligned.action.transformed_variables.get(self.variable, self.variable),
                    'y': aligned.action.transformed_variables.get(other.variable, other.variable)
                }
            elif isinstance(other, FrozenQuery):
                raise TypeError(f"Cannot compare types {self.__class__} and {other.__class__}")
            else:
                aligned = self.branch.add_data(other)
                inputs = {'x': self.variable, 'y': aligned.current_variables[0]}
        else:  # only one variable
            inputs = {'x': self.variable}
            aligned = self.branch
        newbranch = aligned.operate(string, **inputs)
        return Dissociated(self.handler, newbranch, newbranch.current_variables[-1], self)

    def __add__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} + {y}', other)

    def __sub__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} - {y}', other)

    def __mul__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} * {y}', other)

    def __truediv__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} / {y}', other)

    def __pow__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} ^ {y}', other)

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__
    __rpow__ = __pow__

    def __neg__(self) -> 'Dissociated':
        return self._apply_aligning_func('(-{x})')

    def __abs__(self) -> 'Dissociated':
        return abs(self)

    def __floor__(self) -> 'Dissociated':
        return floor(self)

    def __ceil__(self) -> 'Dissociated':
        return ceil(self)

    def __round__(self, n=None):
        return round(self, n)

    def __eq__(self, other: Union['Dissociated', int, float, str]) -> 'Dissociated':
        string = '{x} = {y}'
        if isinstance(other, str):
            if is_regex(other):
                string = '{x} =~ {y}'
                other = other.strip('/')
        return self._apply_aligning_func(string, other)

    def __ne__(self, other: Union['Dissociated', int, float, str]) -> 'Dissociated':
        string = '{x} <> {y}'
        if isinstance(other, str):
            if is_regex(other):
                string = 'NOT ({x} =~ {y})'
                other = other.strip('/')
        return self._apply_aligning_func(string, other)

    def __lt__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} < {y}', other)

    def __le__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} <= {y}', other)

    def __gt__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} > {y}', other)

    def __ge__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_aligning_func('{x} >= {y}', other)

    def __and__(self, other: 'Dissociated') -> 'Dissociated':
        return self._apply_aligning_func('({x} AND {y})', other)

    def __or__(self, other: 'Dissociated') -> 'Dissociated':
        return self._apply_aligning_func('({x} OR {y})', other)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if len(inputs) != 1:
            raise NotImplementedError(f"Can only translate numpy methods that have 1 argument")
        if method == '__call__':
            return _template_operator(f'{ufunc.__name__}({{x}})', ufunc.__name__, ufunc, self, True, True)
        if method == 'reduce':
            translate = {'add': sum, 'logical_or': any, 'logical_and': all, 'minimum': min, 'maximum': max}[ufunc.__name__]
            return translate(self, wrt=kwargs['axis'])
        else:
            raise NotImplementedError(f"Can only translate numpy methods that are called using __call__ or reduce. Not {method}")

    def _post_process(self, result: py2neo.database.Cursor, squeeze: bool = True) -> Table:
        df = result.to_data_frame()
        assert len(df.columns) == 1
        vs = df.iloc[:, 0].values
        if squeeze and len(vs) == 1:
            return vs[0]
        return vs

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            return query.returns(self.variable)


from weaveio import *
from weaveio.readquery.results import Row
from astropy.table import vstack
import pytest

def test_as_table(data):
    obs = data.obs
    all_query = count(obs.l1single_spectra, wrt=obs)
    all_r = obs[[all_query, 'id']]
    split_obs = split(obs)
    query = count(split_obs.l1single_spectra, wrt=split_obs)
    r = split_obs[[query, 'id']]

    rows = []
    for index, subquery in r:
        row = subquery()
        assert isinstance(row, Row)
        assert np.all(row['id'] == index[0])
        rows.append(row)

    row_table = vstack(rows)
    row_table.sort('id')
    all_table = all_r(limit=None)
    all_table.sort('id')
    assert all(row_table == all_table)


def test_as_attribute(data):
    obs = data.obs
    split_obs = split(obs)
    query = count(split_obs.l1single_spectra, wrt=split_obs)
    all_query = count(obs.l1single_spectra, wrt=obs)

    rows = []
    for index, q in query:
        rows.append(q())
    assert sorted(all_query(limit=None)) == sorted(rows)


def test_continue_query_func(data):
    obs = data.obs
    split_obs = split(obs)
    query = split_obs.l1single_spectra

    for index, q in query:
        assert count(q)() > (999*2)
        break

def test_continue_query_traverse(data):
    obs = data.obs
    obs = obs[obs.id == 3133]
    split_obs = split(obs)
    query = split_obs.l1stack_spectra

    for index, q in query:
        assert np.all(count(q.l1single_spectra, wrt=q)() == 3)


from dataclasses import dataclass
from string import ascii_letters, ascii_uppercase, digits
from typing import Dict, List, Any

import hypothesis.strategies as st
from hypothesis import given, note, settings
from hypothesis.strategies import builds, data, composite
from pytest import mark

from weaveio.graph import _convert_datatypes, is_null


@composite
def _key_strategy(draw, start=st.text(min_size=1, max_size=1, alphabet=ascii_letters),
                 tail=st.text(min_size=0, alphabet=ascii_letters+digits)):
    return draw(start) + draw(tail)
key_strategy = _key_strategy()


@composite
def _label_strategy(draw, start=st.text(min_size=1, max_size=1, alphabet=ascii_uppercase),
                 tail=st.text(min_size=1, alphabet=ascii_letters+digits)):
    return draw(start) + draw(tail)
label_strategy = _label_strategy()

labels_strategy = st.lists(label_strategy, min_size=1, max_size=5, unique=True)

_basic_types = [st.integers(min_value=-9999, max_value=999), st.text(alphabet=ascii_letters+digits), st.floats(allow_nan=False, allow_infinity=False)]
_nan_basic_types = [st.integers(min_value=-9999, max_value=999), st.text(alphabet=ascii_letters+digits), st.floats(allow_nan=True, allow_infinity=False)]
_list_types = [st.lists(x, max_size=10) for x in _basic_types]  # limited to prevent long waits
_nan_list_types = [st.lists(x, max_size=10) for x in _nan_basic_types]  # limited to prevent long waits

baseproperty_strategy = st.one_of(*_basic_types)
neo4jproperty_strategy = st.one_of(*_nan_basic_types+_nan_list_types)
properties_strategy = st.dictionaries(key_strategy, neo4jproperty_strategy)
identproperties_strategy = st.dictionaries(key_strategy, baseproperty_strategy)


@dataclass
class Node:
    labels: List[str]
    properties: Dict[str, Any]
    identproperties: Dict[str, Any]

    @property
    def allproperties(self):
        props = self.properties.copy()
        props.update(self.identproperties)
        return props

    def __eq__(self, other):
        return set(self.labels) == set(other.labels) and is_equal(self.allproperties, other.allproperties)


@dataclass
class Relation:
    reltype: str
    properties: Dict[str, Any]
    identproperties: Dict[str, Any]

    @property
    def allproperties(self):
        props = self.properties.copy()
        props.update(self.identproperties)
        return props


node_strategy = builds(Node, labels=labels_strategy, properties=properties_strategy, identproperties=identproperties_strategy)
rel_strategy = builds(Relation, reltype=label_strategy, properties=properties_strategy, identproperties=identproperties_strategy)


def labels_from_labels(labels: List, how, sampler):
    if how == False:
        return labels
    elif how == 'entire':  # must be entirely different
        return sampler.draw(labels_strategy.filter(lambda xs: not any(x in labels for x in xs)))
    elif how == 'crop':  # return the same list up until an index
        if len(labels) == 1:
            return labels  # odd, but we need to ignore it here
        index = sampler.draw(st.integers(min_value=1, max_value=len(labels) - 1))
        return labels[:index]
    elif how == 'extend':  # just add more labels
        return labels + sampler.draw(labels_strategy.filter(lambda xs: not any(x in labels for x in xs)
                                                                       and len(xs) > 1))
    elif how == 'crop&extend':
        if len(labels) == 1:
            return labels
        return labels[:-1] + [sampler.draw(label_strategy.filter(lambda x: x not in labels[:-1]))]
    else:
        raise ValueError(f"Method to change labels {how} is unknown")


def properties_from_properties(properties: Dict, how, sampler, properties_strategy_type, property_strategy_type, exclude_keys=None):
    if exclude_keys is None:
        exclude_keys = []
    exclude_keys = exclude_keys.copy()
    new = properties.copy()
    keys = list(properties.keys())
    if how == False:
        return properties
    elif how == 'entirekeys':  # must be entirely different
        exclude_keys += list(properties.keys())
        return sampler.draw(properties_strategy_type.filter(lambda d: not any(k in exclude_keys for k in d)))
    elif how == 'crop':  # return the same list up until an index
        if len(properties) == 0:
            return {}  # this situation is handled by 'extend'
        indexes = sampler.draw(st.lists(st.integers(min_value=0, max_value=len(keys) - 1), unique=True, min_size=1, max_size=len(keys)))
        for i in indexes:
            del new[keys[i]]
        return new
    elif how == 'addkeys':  # just add more labels
        exclude_keys += list(properties.keys())
        strat = properties_strategy_type.filter(lambda x: not any(k in exclude_keys for k in x) and len(x))
        additions = sampler.draw(strat)
        new.update(additions)
        return new
    elif how == 'overwritekeys':
        if len(properties) == 0:
            return {}  # this situation is handled by 'extend'
        n = sampler.draw(st.integers(min_value=1, max_value=len(properties)))
        whichkeys = sampler.draw(st.lists(st.integers(min_value=0, max_value=len(keys)-1), min_size=n, max_size=n, unique=True))
        newkeys = [keys[i] for i in whichkeys]
        oldvalues = [properties[k] for k in newkeys]
        newvalues = sampler.draw(st.lists(property_strategy_type.filter(lambda x: x not in oldvalues),
                                          min_size=n, max_size=n)
                                 )
        for k, v in zip(newkeys, newvalues):
            new[k] = v
        return new
    else:
        raise ValueError(f"Method to change properties {how} is unknown")

def create_node_from_node(node, sampler, different_labels, different_properties, different_identproperties):
    new_identproperties = properties_from_properties(node.identproperties, different_identproperties,
                                                     sampler, identproperties_strategy, baseproperty_strategy)
    new_properties = properties_from_properties(node.properties, different_properties, sampler,
                                                properties_strategy, neo4jproperty_strategy,
                                                list(new_identproperties.keys()))
    new_labels = labels_from_labels(node.labels, different_labels, sampler)
    return Node(new_labels, identproperties=new_identproperties, properties=new_properties)


def create_rel_from_rel(rel, sampler, different_labels, different_properties, different_identproperties):
    new_identproperties = properties_from_properties(rel.identproperties, different_identproperties,
                                                     sampler, identproperties_strategy, baseproperty_strategy)
    new_properties = properties_from_properties(rel.properties, different_properties, sampler,
                                                properties_strategy, neo4jproperty_strategy,
                                                list(new_identproperties.keys()))
    new_reltype = labels_from_labels([rel.reltype], different_labels, sampler)[0]
    return Relation(new_reltype, identproperties=new_identproperties, properties=new_properties)



@given(labels=labels_strategy.filter(lambda x: len(x) > 1), sampler=data())
@mark.parametrize('different_labels', [False, 'entire', 'crop', 'extend', 'crop&extend'])
def test_labels_from_labels(labels, different_labels, sampler):
    newlabels = labels_from_labels(labels, different_labels, sampler)
    note(f"newlabels = {newlabels}")
    note(f"oldlabels = {labels}")
    assert len(newlabels) > 0
    assert len(set(newlabels)) == len(newlabels)
    if different_labels == False:
        assert newlabels == labels
    elif different_labels == 'entire':
        assert not any(l in labels for l in newlabels)
    elif different_labels == 'crop':
        assert newlabels == labels[:len(newlabels)]
    elif different_labels == 'extend':
        assert newlabels[:len(labels)] == labels
        assert len(newlabels) > len(labels)
    elif different_labels == 'crop&extend':
        assert newlabels[0] == labels[0]  # at least one is the same


def is_equal(x, y):
    x = _convert_datatypes(x)
    y = _convert_datatypes(y)
    return x == y or x is y


all_types_dict = {int: st.integers(), float: st.floats(allow_nan=True, allow_infinity=True),
                  str: st.text(), bool: st.booleans(), type(None): st.none(),
                  dict: st.dictionaries(st.text(), neo4jproperty_strategy | st.lists(neo4jproperty_strategy))}

def alter(x, sampler):
    if isinstance(x, (tuple, list)):
        if len(x) > 0:
            n = sampler.draw(st.integers(0, len(x)-1))
            other = sampler.draw(all_types_dict[type(x[n])])
            y = x.copy()
            y[n] = other
            return x, y
        else:
            other = sampler.draw(neo4jproperty_strategy)
            return x, [other]
    other = sampler.draw(all_types_dict[type(x)])
    return x, other


@settings(max_examples=1000)
@given(a=neo4jproperty_strategy)
def test_is_equal(a):
    assert is_equal(a, a)


@settings(max_examples=1000)
@given(a=neo4jproperty_strategy, sampler=st.data())
def test_is_not_equal(a, sampler):
    b = alter(a, sampler)
    note(f'a is {a}')
    note(f'b is {b}')
    assert not is_equal(a, b)



@settings(print_blob=True)
@given(properties=properties_strategy, sampler=data())
@mark.parametrize('different_properties', [False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys'])
def test_properties_from_properties(properties, different_properties, sampler):
    excluded = ['a', 'b']
    newprops = properties_from_properties(properties, different_properties, sampler,
                                          properties_strategy, neo4jproperty_strategy, excluded)
    note(f"newprops = {newprops}")
    note(f"oldprops = {properties}")
    if different_properties == False:
        assert properties == newprops
    elif different_properties == 'entirekeys':
        assert not any(k in properties for k in newprops)
        assert not any(e in newprops for e in excluded)
    elif different_properties == 'crop':
        assert len(newprops) < len(properties) or len(properties) == 0
        assert all(is_equal(properties[k], v) for k, v in newprops.items())
    elif different_properties == 'addkeys':
        assert all(is_equal(newprops[k], v) for k, v in properties.items())
        assert len(newprops) > len(properties)
    elif different_properties == 'overwritekeys':
        assert len(newprops) == len(properties)
        if len(properties) > 0:
            # at least one different
            assert any(not is_equal(newprops[k], v) for k, v in properties.items())


@given(x=neo4jproperty_strategy)
def test_convert(x):
    y = _convert_datatypes(x)
    if isinstance(x, (tuple, list)):
        assert all(isinstance(yi, type(y[0])) for yi in y), "Not a homogeneous list"
        assert not any(is_null(yi) for yi in y), "List contained nulls"
    else:
        assert not is_null(y)



import pytest
from weaveio.basequery.hierarchy import HomogeneousHierarchyFrozenQuery, SingleHierarchyFrozenQuery, \
    HeterogeneousHierarchyFrozenQuery, IdentifiedHomogeneousHierarchyFrozenQuery
from weaveio.basequery.query import AmbiguousPathError
from weaveio.basequery.query_objects import Node, Generator, Condition
from weaveio.basequery.tests.example_structures.one2one import HierarchyA, MyDataOne2One


hierarchies = MyDataOne2One('.').hierarchies

def test_begin_with_heterogeneous(data_one2one):
    hetero = data_one2one.handler.begin_with_heterogeneous()
    assert isinstance(hetero, HeterogeneousHierarchyFrozenQuery)


@pytest.mark.parametrize('hier', hierarchies)
def test_get_homogeneous_from_data(data_one2one, hier):
    hierarchies = data_one2one.__getattr__(hier.plural_name)
    assert isinstance(hierarchies, HomogeneousHierarchyFrozenQuery)
    assert hierarchies.hierarchy_type is hier
    query = hierarchies.branch
    assert not query.predicates
    assert not query.returns
    assert not query.exist_branches
    assert not query.conditions
    assert len(query.matches) == 1 and query.matches[0][-1].name == f'{hier.singular_name}0'


@pytest.mark.parametrize('hier', hierarchies)
def test_return_homogeneous_with_indexer(data_one2one, hier):
    hierarchies = data_one2one.__getattr__(hier.plural_name)
    query = hierarchies._prepare_query()
    indexer = Node(name='none0')
    assert query.returns == [query.current_node, indexer]


def test_index_by_single_identifier(data_one2one):
    single = data_one2one.hierarchyas['1']
    query = single.query
    assert isinstance(single, SingleHierarchyFrozenQuery)
    assert single.hierarchy_type is HierarchyA
    assert not query.predicates
    assert not query.returns
    assert not query.exist_branches
    assert len(query.matches) == 1 and query.matches[0].nodes[-1].name == 'hierarchya0'
    assert query.conditions == Condition(query.matches[0].nodes[-1].id, '=', '1')


def test_index_a_plural_by_single_identifier_fails(data_one2one):
    with pytest.raises(AmbiguousPathError):
        data_one2one.hierarchyas.hierarchybs['1']


def test_index_a_plural_by_list_of_identifiers_succeeds(data_one2one):
    data_one2one.hierarchyas.hierarchybs[['1', '2']]


def test_get_homogeneous_from_homogeneous(data_one2one):
    homo = data_one2one.hierarchyas.hierarchybs
    assert isinstance(homo, HomogeneousHierarchyFrozenQuery)
    assert len(homo.branch.matches) == 2


def test_get_single_from_homogeneous_invalid(data_one2one):
    with pytest.raises(AmbiguousPathError):
        data_one2one.hierarchyas.hierarchyb


def test_get_single_from_single_ascending(data_one2one):
    single = data_one2one.hierarchyas['1'].hierarchyb
    assert isinstance(single, SingleHierarchyFrozenQuery)


def test_index_by_multiple_identifiers(data_one2one):
    names = ['2.fits', '1.fits']
    multi = data_one2one.hierarchyas[names]
    assert isinstance(multi, IdentifiedHomogeneousHierarchyFrozenQuery)
    assert multi._identifiers == ['2.fits', '1.fits']


def test_ambiguous_hierarchy_names_must_be_explicit(data_one2one):
    pass


from typing import Callable, TYPE_CHECKING

import numpy as np

from .utilities import mask_infs
from .base import BaseQuery
if TYPE_CHECKING:
    pass

__all__ = ['sum', 'max', 'min', 'mean', 'std', 'count', 'any', 'all', 'exists', 'array']

python_any = any
python_all = all
python_max = max
python_min = min
python_sum = sum


def _template_aggregator(string_op, predicate, python_func: Callable, item: BaseQuery, wrt: BaseQuery = None,
                         remove_infs: bool = True, expected_dtype: str = None, returns_dtype:str = None,
                         op_name=None, distinct=False, nulls=False, args=None, kwargs=None):
    from ..data import Data
    from .objects import AttributeQuery
    if isinstance(wrt, Data):
        wrt = None
    elif isinstance(wrt, AttributeQuery):
        raise TypeError(f"Cannot aggregate {item} with respect to an attribute {wrt}. You can only aggregate with respect to an object.")
    if op_name is None:
        op_name = string_op
    try:
        return item._aggregate(wrt, string_op, op_name, predicate, expected_dtype, returns_dtype, remove_infs, distinct, nulls=nulls)
    except AttributeError:
        if distinct:
            item = set(item)
        return python_func(item, *args, **kwargs)


def sum(item, wrt=None, *args, **kwargs):
    return _template_aggregator('sum', False, python_sum, item, wrt, expected_dtype='number', args=args, kwargs=kwargs)


def max(item, wrt=None, *args, **kwargs):
    return _template_aggregator('max', False, python_max, item, wrt, expected_dtype='number', args=args, kwargs=kwargs)


def min(item, wrt=None, *args, **kwargs):
    return _template_aggregator('min', False, python_min, item, wrt, expected_dtype='number', args=args, kwargs=kwargs)


def count(item, wrt=None, *args, **kwargs):
    return _template_aggregator('count', False, len, item, wrt, returns_dtype='integer', args=args, kwargs=kwargs)


def exists(item, wrt=None, *args, **kwargs):
    return _template_aggregator('count', False, len, item, wrt, returns_dtype='number', args=args, kwargs=kwargs) > 0


def std(item, wrt=None, *args, **kwargs):
    return _template_aggregator('stDev', False, np.std, item, wrt, expected_dtype='number', args=args, kwargs=kwargs)


def mean(item, wrt=None, *args, **kwargs):
    return _template_aggregator('avg', False, np.mean, item, wrt, expected_dtype='number', op_name='mean',
                                args=args, kwargs=kwargs)

#predicates

def all(item, wrt=None, *args, **kwargs):
    return _template_aggregator('all', True, python_all, item, wrt, args=args, returns_dtype='boolean', expected_dtype='boolean', kwargs=kwargs)


def any(item, wrt=None, *args, **kwargs):
    return _template_aggregator('any', True, python_any, item, wrt, args=args, returns_dtype='boolean', expected_dtype='boolean', kwargs=kwargs)


def array(item, wrt=None, distinct=False, nulls=False, *args, **kwargs):
    return _template_aggregator('collect', False, list, item, wrt, distinct=distinct, nulls=nulls, args=args, kwargs=kwargs)


# Copied the Context structure from Pymc3
import threading
from sys import modules
from typing import Optional, Type, List, Union, TypeVar


T = TypeVar("T", bound="ContextMeta")


class ContextError(Exception):
    pass


class ContextMeta(type):
    """Functionality for objects that put themselves in a context using
    the `with` statement.
    """

    def __new__(cls, name, bases, dct, **kargs):  # pylint: disable=unused-argument
        "Add __enter__ and __exit__ methods to the class."

        def __enter__(self):
            self.__class__.context_class.get_contexts().append(self)
            return self

        def __exit__(self, typ, value, traceback):  # pylint: disable=unused-argument
            self.__class__.context_class.get_contexts().pop()

        dct[__enter__.__name__] = __enter__
        dct[__exit__.__name__] = __exit__

        # We strip off keyword args, per the warning from
        # StackExchange:
        # DO NOT send "**kargs" to "type.__new__".  It won't catch them and
        # you'll get a "TypeError: type() takes 1 or 3 arguments" exception.
        return super().__new__(cls, name, bases, dct)

    # FIXME: is there a more elegant way to automatically add methods to the class that
    # are instance methods instead of class methods?
    def __init__(
        cls, name, bases, nmspc, context_class: Optional[Type] = None, **kwargs
    ):  # pylint: disable=unused-argument
        """Add ``__enter__`` and ``__exit__`` methods to the new class automatically."""
        if context_class is not None:
            cls._context_class = context_class
        super().__init__(name, bases, nmspc)

    def get_context(cls, error_if_none=True) -> Optional['Graph']:
        """Return the most recently pushed context object of type ``cls``
        on the stack, or ``None``. If ``error_if_none`` is True (default),
        raise a ``ContextError`` instead of returning ``None``."""
        idx = -1
        while True:
            try:
                candidate = cls.get_contexts()[idx]  # type: Optional[T]
            except IndexError as e:
                # Calling code expects to get a TypeError if the entity
                # is unfound, and there's too much to fix.
                if error_if_none:
                    raise ContextError("No %s on context stack" % str(cls))
                return None
            return candidate


    def get_contexts(cls) -> List[T]:
        """Return a stack of context instances for the ``context_class``
        of ``cls``."""
        # This lazily creates the context class's contexts
        # thread-local object, as needed. This seems inelegant to me,
        # but since the context class is not guaranteed to exist when
        # the metaclass is being instantiated, I couldn't figure out a
        # better way. [2019/10/11:rpg]

        # no race-condition here, contexts is a thread-local object
        # be sure not to override contexts in a subclass however!
        context_class = cls.context_class
        assert isinstance(context_class, type), (
            "Name of context class, %s was not resolvable to a class" % context_class
        )
        if not hasattr(context_class, "contexts"):
            context_class.contexts = threading.local()

        contexts = context_class.contexts

        if not hasattr(contexts, "stack"):
            contexts.stack = []
        return contexts.stack

    # the following complex property accessor is necessary because the
    # context_class may not have been created at the point it is
    # specified, so the context_class may be a class *name* rather
    # than a class.
    @property
    def context_class(cls) -> Type:
        def resolve_type(c: Union[Type, str]) -> Type:
            if isinstance(c, str):
                c = getattr(modules[cls.__module__], c)
            if isinstance(c, type):
                return c
            raise ValueError("Cannot resolve context class %s" % c)

        assert cls is not None
        if isinstance(cls._context_class, str):
            cls._context_class = resolve_type(cls._context_class)
        if not isinstance(cls._context_class, (str, type)):
            raise ValueError(
                "Context class for %s, %s, is not of the right type"
                % (cls.__name__, cls._context_class)
            )
        return cls._context_class

    # Inherit context class from parent
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.context_class = super().context_class

    # Initialize object in its own context...
    # Merged from InitContextMeta in the original.
    def __call__(cls, *args, **kwargs):
        instance = cls.__new__(cls, *args, **kwargs)
        with instance:  # appends context
            instance.__init__(*args, **kwargs)
        return instance



import pytest
import numpy as np
from astropy.table import Table

from weaveio.basequery.tests.example_structures.one2one import HierarchyA, HierarchyB


def test_instantiate_db(database_one2one):
    pass


def test_return_homogeneous(database_one2one):
    for h in database_one2one.hierarchyas():
        assert isinstance(h, HierarchyA)
        assert h.a_factor_a == 'a'
        assert h.a_factor_b == 'b'


def test_return_single(database_one2one):
    b = database_one2one.hierarchyas['1.fits'].hierarchyb()
    assert isinstance(b, HierarchyB)
    assert b.otherid == '1.fits'


def test_empty_id_raise_keyerror(database_one2one):
    with pytest.raises(KeyError, match="nonexistent_id"):
        database_one2one.hierarchyas['nonexistent_id'].hierarchyb()


def test_uchained_queries(database_one2one):
    first = database_one2one.hierarchyas
    second = first.hierarchybs
    first()
    second()


def test_multiple_ids(database_one2one):
    names = ['1.fits', '2.fits', '1.fits']
    a = database_one2one.hierarchyas[names]()
    assert [i.id for i in a] == names


def test_multiple_ids_keyerror(database_one2one):
    names = ['1.fits', '2.fits', 'nan']
    with pytest.raises(KeyError, match='nan'):
        database_one2one.hierarchyas[names]()


def test_single_factor_is_scalar(database_one2one):
    assert database_one2one.hierarchyas['1.fits'].a_factor_a() == 'a'


def test_column_factor_is_vector(database_one2one):
    np.testing.assert_array_equal(database_one2one.hierarchyas['1.fits', '2.fits'].a_factor_as(), ['a', 'a'])


def convert_object_array_of_lists2_array(array):
    if array.dtype == object and not isinstance(array.ravel()[0], str):
        shape = array.shape
        return np.reshape([np.asarray(i) for i in array.ravel()], shape + (-1,))
    else:
        return array

@pytest.mark.parametrize('columns,colshape,is_unstructured',
                         (
                                 ['c_factor_as', (2, ), False], [['c_factor_as'], (2, ), False],
                                 ['a_factor_a', tuple(), False], [['a_factor_a'], tuple(), False],
                                 ['f_factor_as', (1, ), True], [['f_factor_as'], (1, ), True]
                         ),
                          ids=["['c_factor_as']", "[['c_factor_as']]",
                               "['a_factor_a']", "[['a_factor_a']]",
                               "['f_factor_as']", "[['f_factor_as']]"])
@pytest.mark.parametrize('idfilter,idshape', ([None, (5,)], [('1.fits', ), (1, )], [['1.fits'], (1,)]),
                         ids=["", "['1.fits']", "[['1.fits']]"])
def test_table_return_shape(database_one2one, columns, is_unstructured, idfilter, idshape, colshape):
    """
    Test that [[colname]] type getitems always return astropy tables with the correct shape,
    plural colnames should make a list structure within it.
    """
    parent = database_one2one.hierarchyas
    if idfilter is not None:
        parent = parent.__getitem__(idfilter)
    structure = parent.__getitem__(columns)
    result = structure()
    if isinstance(columns, list):
        assert isinstance(result, Table)
        result = result[result.colnames[0]].data
    else:
        assert isinstance(result, np.ndarray)
    if is_unstructured:
        assert isinstance(result[0], list)
    result = convert_object_array_of_lists2_array(result)
    expected = np.empty(idshape + colshape, dtype=str)
    expected[:] = 'a'
    np.testing.assert_array_equal(result, expected)


def test_names_with_dots_resolve_correctly(database_one2one):
    # cypher doesn't allow dots in its names, but we want the dots in our table
    table = database_one2one.hierarchyas[['hierarchyd.shared_factor_name']]()
    assert table.colnames == ['hierarchyd.shared_factor_name']




from functools import reduce

from .wrappers import QueryWrapper


class AlignedQuery(QueryWrapper):

    @property
    def aligned(self):
        return self.queries

    def _get_wrt(self):
        d = {}
        previous_nodes = []
        for q in self.queries.values():
            _previous_nodes = []
            while True:
                d[q._node] = q
                _previous_nodes.append(q._node)
                q = q._previous
                if q is None:
                    break
            previous_nodes.append(_previous_nodes)
        return d[max(reduce(set.intersection, map(set, previous_nodes)))]

    def _precompile(self):
        from .objects import TableQuery
        prefix, suffix = '', ''
        if any(isinstance(v, TableQuery) for v in self.queries.values()):
            if not all(isinstance(v, TableQuery) for v in self.queries.values()):
                raise ValueError('All queries must be tables if any are tables')
            suffix = '_'
        wrt = self._get_wrt()
        return wrt[{f"{prefix}{k}{suffix}": q for k, q in self.queries.items()}]


def align(*unnamed_queries, **named_queries):
    from .objects import BaseQuery
    queries = unnamed_queries + tuple(named_queries.values())
    if not all(isinstance(q, BaseQuery) for q in queries):
        raise TypeError('All queries must be Queries objects')
    data = queries[0]._data
    if not all(q._data is data for q in queries):
        raise ValueError('All queries must be from the same parent Data object')
    return AlignedQuery(*unnamed_queries, **named_queries)




from astropy.io import fits
from astropy.table import Table
import pytest
from weaveio import *

@pytest.fixture(scope='module')
def nsky_expected(data):
    return {sum(Table(fits.open(data.rootdir / p)['FIBTABLE'].data)['TARGUSE'] == 'S') for p in data.raw_files.path()}


def test_example1a(data):
    runid = 1003453
    nsky = sum(data.runs[runid].targuses == 'S')()
    q = data.runs[runid].raw_file.path
    t = Table(fits.open(data.rootdir / q()[0])['FIBTABLE'].data)
    assert nsky == sum(t['TARGUSE'] == 'S')

def test_example1b(data, nsky_expected):
    nsky = set(sum(data.runs.targuses == 'S', wrt=data.runs)())  # sum the number of skytargets with respect to their runs
    assert nsky == nsky_expected


def test_example1c(data, nsky_expected):
    nsky = sum(data.runs.targuses == 'S', wrt=data.runs)  # sum the number of skytargets with respect to their runs
    query_table = data.runs[['id', nsky]]  # design a table by using the square brackets
    concrete_table = query_table()
    expected_nruns = count(data.runs)()
    assert len(concrete_table) == expected_nruns
    assert set(concrete_table.sum) == nsky_expected
    assert len(set(concrete_table.id)) == expected_nruns


def test_example2(data):
    yesterday = 57811
    runs = data.runs
    is_red = runs.camera == 'red'
    is_yesterday = floor(runs.exposure.mjd) == yesterday  # round to integer, which is the day
    runs = runs[is_red & is_yesterday]  # filter the runs first
    spectra = runs.l1single_spectra
    sky_spectra = spectra[spectra.targuse == 'S']
    table = sky_spectra[['wvl', 'flux']]
    assert count(table)() == 690
    assert not any(np.any(t['wvl'].mask) for t in table(limit=10))  # just check the first few


def test_example3(data):
    l2s = data.l2stacks
    l2s = l2s[(l2s.ob.mjd >= 57780) & any(l2s.fibre_target.surveys == '/WL.*/', wrt=l2s.fibre_target)]
    l2s = l2s[l2s['ha_6562.80_flux'] > 0]
    table = l2s[['ha_6562.80_flux', 'z']]()  # type: Table
    assert len(table) == 914
    assert table['z'].min(), table['z'].max() == (-0.004380459951814368, 0.9470194571839268)
    assert table['ha_6562.80_flux'].min(), table['ha_6562.80_flux'].max() == (0.013978097400486398, 2.736550702489109e+18)


def test_example4(data):
    from astropy.table import Table
    import weaveio
    fname = Path(weaveio.__file__).parents[0] / 'tests/my_table.ascii'
    table = Table.read(fname, format='ascii')
    rows, targets = join(table, 'cname', data.weave_targets)
    mjds = targets.exposures.mjd  # get the mjd of the plate exposures for each target
    q = targets['cname', rows['modelMag_i'], {'mjds': mjds, 'nobservations': count(mjds, wrt=targets)}]
    t = q()
    assert len(t) == len(table)
    assert np.all(np.sort(t['cname']) == np.sort(table['cname']))
    assert np.all(t['nobservations'] >= 3)


def test_adjunct(data):
    t = data.l1single_spectra[data.l1single_spectra.camera == 'red']
    t = t[['camera', t.adjunct.camera]](limit=5)
    assert np.all(t['camera0'] == 'red')
    assert np.all(t['camera1'] == 'blue')




import pytest

from weaveio.basequery.factor import SingleFactorFrozenQuery, ColumnFactorFrozenQuery, RowFactorFrozenQuery, TableFactorFrozenQuery
from weaveio.basequery.hierarchy import HomogeneousHierarchyFrozenQuery
from weaveio.basequery.query import AmbiguousPathError
from weaveio.basequery.query_objects import Node, NodeProperty
from weaveio.basequery.tests.example_structures.one2one import HierarchyA, HierarchyB, HierarchyC
from weaveio.utilities import quote


def test_single_hierarchy_direct_single_factor(data_one2one):
    """get a single factor from the parent hierarchy directly"""
    single = data_one2one.hierarchyas['1'].a_factor_a
    query = single.query
    assert isinstance(single, SingleFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyA', name='hierarchya0'), 'a_factor_a')
    assert len(query.returns) == 1   # a_factor_a and no indexer


def test_single_hierarchy_indirect_single_factor(data_one2one):
    """get a single factor from a hierarchy above the parent"""
    single = data_one2one.hierarchyas['1'].b_factor_a
    query = single.query
    assert isinstance(single, SingleFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyB', name='hierarchyb0'), 'b_factor_a')
    assert len(query.returns) == 1   # b_factor_b and no indexer


def test_single_hierarchy_fails_with_unknown_name(data_one2one):
    with pytest.raises(AttributeError):
        data_one2one.hierarchyas['1'].unknown


def test_homogeneous_hierarchy_fails_with_unknown_name(data_one2one):
    with pytest.raises(AttributeError):
        data_one2one.hierarchyas.unknowns


def test_homogeneous_hierarchy_direct_plural_factor(data_one2one):
    single = data_one2one.hierarchyas.a_factor_as
    query = single.query
    assert isinstance(single, ColumnFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyA', name='hierarchya0'), 'a_factor_a')
    assert len(query.returns) == 1   # a_factor_a and no indexer


def test_homogeneous_hierarchy_indirect_plural_factor(data_one2one):
    single = data_one2one.hierarchyas.b_factor_as
    query = single.query
    assert isinstance(single, ColumnFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyB', name='hierarchyb0'), 'b_factor_a')
    assert len(query.returns) == 1   # b_factor_b and no indexer


def test_identified_homogeneous_hierarchy_fails_with_unknown_name(data_one2one):
    with pytest.raises(AttributeError):
        data_one2one.hierarchyas[['1', '2']].unknowns


def test_identified_homogeneous_hierarchy_direct_plural_factor(data_one2one):
    single = data_one2one.hierarchyas[['1', '2']].a_factor_as
    query = single.query
    assert isinstance(single, ColumnFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyA', name='hierarchya0'), 'a_factor_a')
    assert len(query.returns) == 1   # a_factor_a and no indexer


def test_identified_homogeneous_hierarchy_indirect_plural_factor(data_one2one):
    single = data_one2one.hierarchyas[['1', '2']].b_factor_as
    query = single.query
    assert isinstance(single, ColumnFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyB', name='hierarchyb0'), 'b_factor_a')
    assert len(query.returns) == 1   # b_factor_b and no indexer


def test_heterogeneous_plural_factor(data_one2one):
    factors = data_one2one.b_factor_as
    query = factors.query
    assert isinstance(factors, ColumnFactorFrozenQuery)
    assert query.returns[0] == NodeProperty(Node(label='HierarchyB', name='hierarchyb0'), 'b_factor_a')
    assert len(query.returns) == 1  # b_factor_b and no indexer


@pytest.mark.parametrize('typ', [tuple, list])
@pytest.mark.parametrize('hiers', [['a'], ['b'], ['a', 'b']])
def test_single_hierarchy_row_of_factors(data_one2one, typ, hiers):
    items, hiers = zip(*[(item, h) for h in hiers for item in [f'{h}_factor_{i}' for i in 'ab']])
    items = typ(items)
    row = data_one2one.hierarchyas['1'].__getitem__(items)
    assert isinstance(row, RowFactorFrozenQuery)
    for i, (item, hier) in enumerate(zip(items, hiers)):
        prop = row.branch.returns[i]
        assert prop.property_name == item
        assert prop.node.label == f'Hierarchy{hier.upper()}'
    if typ is list:
        assert row.return_keys == items
    elif typ is tuple:
        assert row.return_keys is None
    else:
        assert False, "Bad arguments"


@pytest.mark.parametrize('make_plural_name', [True, False])
@pytest.mark.parametrize('factor_name', ['a', 'b'])
@pytest.mark.parametrize('hierarchy,multiple_from_a', ([HierarchyB, False], [HierarchyC, True]))
def test_plural_factor_name_required_by_getattr(data_one2one, hierarchy, multiple_from_a, factor_name, make_plural_name):
    """
    When selecting factors by getattr, plurality is required when there will be more than one result returned
    We tests this by starting at hierarchya and selecting the factors of the parent hierarchies.
    """
    factor_name = hierarchy.__name__[-1].lower() + '_factor_' + factor_name
    name = factor_name + 's' if make_plural_name else factor_name
    parent = data_one2one.hierarchyas
    if make_plural_name:
        assert isinstance(parent.__getattr__(name), ColumnFactorFrozenQuery)
    else:
        with pytest.raises(AmbiguousPathError):
            parent.__getattr__(name)


@pytest.mark.parametrize('make_plural_name', [True, False], ids=lambda v: "s']" if v else "']")
@pytest.mark.parametrize('factor_name', ['a', 'b'], ids=lambda v: f"{v}")
@pytest.mark.parametrize('hierarchy,multiple_from_a', ([HierarchyB, False], [HierarchyC, True]), ids=["['b_factor_", "['c_factor_"])
def test_plural_factor_name_required_by_getitem(data_one2one, hierarchy, multiple_from_a, factor_name, make_plural_name):
    """
    When selecting factors by getitem, plurality is required when each item in the hierarchy before will have multiple items
    We tests this by starting at hierarchya and selecting the factors of the parent hierarchies.
    """
    factor_name = hierarchy.__name__[-1].lower() + '_factor_' + factor_name
    name = factor_name + 's' if make_plural_name else factor_name
    parent = data_one2one.hierarchyas
    if make_plural_name:  # requesting a plural is always allowed
        assert isinstance(parent.__getitem__(name), ColumnFactorFrozenQuery)
    else:
        if multiple_from_a:
            with pytest.raises(AmbiguousPathError):
                parent.__getitem__(name)
        else:
            assert isinstance(parent.__getitem__(name), ColumnFactorFrozenQuery)



@pytest.mark.parametrize('hiers', [['a'], ['b'], ['a', 'b']])
@pytest.mark.parametrize('factor_intype', [tuple, list])
@pytest.mark.parametrize('factor_names', [['a'], ['b'], ['a', 'b']], ids=lambda x: str(x))
@pytest.mark.parametrize('idfilter', ['1', ['1', '2'], ('1', '2'), None],
ids=lambda v: f'hierarchies[{quote(v)}]'.replace('(', '').replace(')', '') if v is not None else 'hierarchies')
def test_tablelike_factors_by_getitem(data_one2one, factor_intype, hiers, factor_names, idfilter):
    """
    Selecting factors by __getitem__ can yield:
        a whole factor_table for multiple hierarchies and multiple items
        a factor_row for one hierarchy and multiple items
        a factor_column for multiple hierarchies and one factor
        a single factor for one hierarchy and one factor

    In getitem, plurality is done on a "foreach" basis, so when the individual hierarchy has more than one value for a factor_name,
    it is pluralised.
    E.g.
    The hierarchy type Run has multiple values for `cname` so `cname` must become `cnames` in any __getitem__ query
        runs[id1, id2]['cnames']
    The hierarchy type Run has only has one value for `expmjd` so `expmjd` can be `expmjd` or `expmjds` depending on what you want
        runs[id1, id2]['expmjd']
        runs[id1, id2]['expmjds']
    """
    items, hiers = zip(*[(item, h) for h in hiers for item in [f'{h}_factor_{i}' for i in factor_names]])
    items = factor_intype(items)

    structure = data_one2one.hierarchyas
    if idfilter is not None:
        structure = structure.__getitem__(idfilter)

    if isinstance(idfilter, (list, tuple)) or idfilter is None:
        if isinstance(items, (list, tuple)):
            querytype = TableFactorFrozenQuery
        else:
            querytype = ColumnFactorFrozenQuery
    else:  # scalar, therefore a single
        if isinstance(items, (list, tuple)):
            querytype = RowFactorFrozenQuery
        else:
            assert False, "bad arguments"  # this is never reached in this test, see below

    table = structure.__getitem__(items)
    assert isinstance(table, querytype), f"data.hierarchyas[{idfilter}][{items}]"

    zippable_items = [items] if not isinstance(items, (tuple, list)) else items
    for i, (item, hier) in enumerate(zip(zippable_items, hiers)):
        prop = table.branch.returns[i]
        assert prop.property_name == item
        assert prop.node.label == f'Hierarchy{hier.upper()}'
    if isinstance(items, list):
        assert table.return_keys == items
    elif isinstance(items, tuple):
        assert table.return_keys is None
    elif isinstance(items, str):
        pass
    else:
        assert False, "Bad arguments"


@pytest.mark.parametrize('hier', ['a', 'b'], ids=lambda v: f"['{v}_factor_a']")
@pytest.mark.parametrize('idfilter', ['1', ['1', '2'], ('1', '2'), None],
ids=lambda v: f'hierarchies[{quote(v)}]'.replace('(', '').replace(')', '') if v is not None else 'hierarchies')
def test_direct_single_factors_by_getitem(data_one2one, idfilter, hier):
    """
    Selecting a single factor_name by getitem is identical to selecting via getattr with the same plurality rules
    """
    if idfilter is not None:
        structure = data_one2one.hierarchyas.__getitem__(idfilter)
    else:
        structure = data_one2one.hierarchyas
    if isinstance(idfilter, (list, tuple)) or idfilter is None:
        querytype = ColumnFactorFrozenQuery
    else:  # scalar, therefore a single
        querytype = SingleFactorFrozenQuery
    factor_name = f'{hier}_factor_a'
    result = structure[factor_name]
    assert isinstance(result, querytype), str(result)
    prop = result.branch.returns[0]
    assert prop.property_name == factor_name
    assert len(result.branch.returns) == 1
    assert prop.node.label == f'Hierarchy{hier.upper()}'


@pytest.mark.parametrize('a_factor_plural', [True, False])
@pytest.mark.parametrize('c_factor_plural', [True, False])
def test_tablelike_factors_by_getitem_raise_when_one_has_wrong_plurality(data_one2one, a_factor_plural, c_factor_plural):
    if a_factor_plural:
        items = ['a_factor_as']
    else:
        items = ['a_factor_a']
    if c_factor_plural:
        items.append('c_factor_as')
    else:
        items.append('c_factor_a')

    if not c_factor_plural:
        with pytest.raises(AmbiguousPathError, match="multiple `c_factor_as`"):
            var = data_one2one.hierarchyas[items]
    else:
        assert isinstance(data_one2one.hierarchyas[items], TableFactorFrozenQuery)


def test_ambiguous_factor_name_raises_error_on_getitem(data_one2one):
    with pytest.raises(AmbiguousPathError):
        var = data_one2one.hierarchyas['1']['shared_factor_name']


def test_ambiguous_factor_name_raises_error_on_getattr(data_one2one):
    with pytest.raises(AmbiguousPathError):
        var = data_one2one.hierarchyas['1'].shared_factor_name


def test_disambiguated_factor_name_on_getitem(data_one2one):
    with pytest.raises(AmbiguousPathError):
        # fails because there is more than one hierarchyc for a hierarchya.
        var = data_one2one.hierarchyas['1']['hierarchyc.shared_factor_name']
    var = data_one2one.hierarchyas['1']['hierarchyd.shared_factor_name']
    var = data_one2one.hierarchyas['1']['hierarchyc.shared_factor_names']


def test_disambiguated_factor_name_on_getattr(data_one2one):
    """Removing ambiguouity from getattr calls is basically just
    another normal getattr call with hierarchies"""
    var = data_one2one.hierarchyas['1'].hierarchyds.shared_factor_names
    var = data_one2one.hierarchyas['1'].hierarchycs.shared_factor_names
    with pytest.raises(AmbiguousPathError):
        # fails because there is more than one hierarchyc for a hierarchya.
        var = data_one2one.hierarchyas['1'].hierarchycs.shared_factor_name


def test_not_plural_raise_ambiguous_path_error(data_one2one):
    with pytest.raises(AmbiguousPathError):
        # hierarchyA has one factor
        # but there are more than one of them, so plural is required
        var = data_one2one.hierarchyas.a_factor_a
    with pytest.raises(AmbiguousPathError):
        # Requesting factors without hierarchies always requires plural
        var = data_one2one.a_factor_a
    with pytest.raises(AmbiguousPathError):
        # hierarchyC has a_factor_a below it, but there is more than one,
        # so it is plural
        var = data_one2one.hierarchycs.a_factor_a



from typing import List


def mask_infs(x):
    return f"CASE WHEN {x} > apoc.math.maxLong() THEN null ELSE {x} END"

def is_regex(other):
    """
    Regex is defined either by:
     1. starting and ending the string with '/'
        OR
     2. When the string contains * and the string doesn't start and end with '"'
    """
    return (other.startswith('/') and other.endswith('/')) or \
           ('*' in other and not (other.startswith('"') and other.endswith('"')))


def remove_successive_duplicate_lines(cypher: List[str]):
    """
    Given a list of cypher statements, remove any lines that appear more than once in succession only
    """
    if not cypher:
        return []
    deduplicated = [cypher[0]]
    for line in cypher[1:]:
        if line != deduplicated[-1]:
            deduplicated.append(line)
    return deduplicated


def safe_name(name):
    if name is None:
        return name
    return '__dot__'.join(name.split('.'))


special_dtypes = {
    ('boolean', 'float'): 'tointeger(tofloat({}))',
    ('boolean', 'number'): 'tointeger({})',
    ('float', 'boolean'): 'toboolean(tointeger({}))',
    ('float', 'number'): '{}',
    ('integer', 'number'): '{}',
    ('number', 'boolean'): 'toboolean(tointeger({}))',
}

def dtype_conversion(from_dtype, to_dtype, string, *replacements):
    if from_dtype == to_dtype:
        return string
    if from_dtype is None or to_dtype is None:
        return string
    func = special_dtypes.get((from_dtype, to_dtype), f'to{to_dtype}({{}})')
    for replacement in replacements:
        string = string.replace(replacement, func.format(replacement))
    return string



from typing import Union

from . import BaseQuery, AttributeQuery
from .objects import ObjectQuery


def split(query: BaseQuery, splitby: Union[AttributeQuery, str, ObjectQuery] = None):
    """
    Splits the given query into unique parts based on the given groupby attribute.
    Each part is treated like a new query and is executed independently.
    This is the weaveio equivalent of an SQL GROUP BY statement.
    Except that you must perform a split *before* the main query.
    This is unlike the SQL GROUP BY statement which is performed *after* the main query.

    Any action you perform on the query after the split is performed on each part automatically.
    Queries that result in duplicates i.e. `data.runs.l1single_spectra.runs` will not be changed, duplications will persist.

    :param query: The query to split.
    :param splitby: The name of the groupby attribute.
    :return: A SplitQuery object containing many queries that were split from the given query.

    Example:
    >>> obs = data.obs
    >>> groups = split(obs, 'id')
    >>> result = groups.l1single_spectra[['flux', 'wvl']]
    >>> result
    <SplitQuery(TableQuery(L1SingleSpectrum-L1SingleSpectrum))>
    for name, group in groups:
    >>>     print(name)
    3901
    >>>     print(group)
    <TableQuery(L1SingleSpectrum-L1SingleSpectrum)>
    >>> print(group())
       flux [14401]     wvl [14401]
    ----------------- ----------------
            0.0 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
       1.203763 .. 0.0 4730.0 .. 5450.0
     0.62065125 .. 0.0 4730.0 .. 5450.0
      347.95587 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
            0.0 .. 0.0 4730.0 .. 5450.0
    """
    if splitby is None and isinstance(query, ObjectQuery):
        try:
            splitby = query._get_default_attr()
        except SyntaxError:
            splitby = query._get_neo4j_id()
    elif isinstance(splitby, str):
        splitby = query[splitby]
    elif isinstance(splitby, ObjectQuery):
        splitby = splitby._get_default_attr()
    elif not isinstance(splitby, AttributeQuery):
        raise TypeError('groupby must be an AttributeQuery, str, or ObjectQuery')
    group_id = query._G.add_groupby(splitby)
    group_eq = splitby._perform_arithmetic(f'{{0}} = {group_id}', '=', group_id, returns_dtype='boolean', parameters=[group_id])
    grouped_query = query._filter_by_mask(group_eq, single=True, split_node=True)
    return grouped_query